# -*- coding: utf-8 -*-
import time

import pytest

from aerospike import exception as e
from aerospike_helpers import expressions as exp
from aerospike_helpers.operations import list_operations
from aerospike_helpers.operations import map_operations
from aerospike_helpers.operations import operations


import aerospike

geo_circle = aerospike.GeoJSON({"type": "AeroCircle", "coordinates": [[-132.0, 37.5], 1000]})

geo_point = aerospike.GeoJSON({"type": "Point", "coordinates": [-132.0, 37.5]})

geo_point2 = aerospike.GeoJSON({"type": "Point", "coordinates": [-60.5, 20]})


class TestPredEveryWhere(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        self.keys = []
        self.test_data = [
            {
                "account_id": j,
                "user_name": "user" + str(j),
                "acct_balance": j * 10,
                "charges": [j + 5, j + 10],
                "meta": {"date": "11/4/2019"},
            }
            for j in range(1, 5)
        ]
        self.test_data.append({"string_list": ["s1", "s2", "s3", "s4"]})
        self.test_data.append({"map_bin": {"k1": 1, "k2": 2, "k3": 3, "k4": 4}})

        georec = {"id": 1, "point": geo_point, "region": geo_circle, "geolist": [geo_point]}

        self.test_data.append(georec)
        self.test_data_bin = "test_data"
        self.keys = [("test", "pred_evry", i + 1) for i, _ in enumerate(self.test_data)]
        # print('self keys is: ', self.keys)

        for key, data in zip(self.keys, self.test_data):
            self.as_connection.put(key, data)

        # cleanup
        yield

        for key in self.keys:
            try:
                self.as_connection.remove(key)
            except e.AerospikeError:
                pass

    @pytest.mark.parametrize(
        "ops, expressions, expected_bins, expected_res, key_num",
        [
            (  # test integer equal
                [operations.increment("account_id", 1)],
                exp.Eq(3, exp.IntBin("account_id")),
                {
                    "account_id": 4,
                    "user_name": "user3",
                    "acct_balance": 30,
                    "charges": [8, 13],
                    "meta": {"date": "11/4/2019"},
                },
                {},
                3,
            ),
            (  # test string equal
                [operations.increment("account_id", 1)],
                exp.Eq("user3", exp.StrBin("user_name")),
                {
                    "account_id": 4,
                    "user_name": "user3",
                    "acct_balance": 30,
                    "charges": [8, 13],
                    "meta": {"date": "11/4/2019"},
                },
                {},
                3,
            ),
            (  # test and
                [
                    list_operations.list_remove_by_index_range("charges", 0, 3, aerospike.LIST_RETURN_COUNT),
                    operations.increment("acct_balance", -23),
                ],
                exp.And(exp.GE(exp.IntBin("acct_balance"), 10), exp.LE(exp.IntBin("acct_balance"), 50)),
                {
                    "account_id": 4,
                    "user_name": "user4",
                    "acct_balance": 17,
                    "charges": [],
                    "meta": {"date": "11/4/2019"},
                },
                {"charges": [0, 1]},
                4,
            ),
            (  # test or
                [map_operations.map_put("meta", "lupdated", "now")],
                exp.Or(exp.Eq(exp.StrBin("user_name"), "user2"), exp.GE(exp.IntBin("acct_balance"), 50)),
                {
                    "account_id": 2,
                    "user_name": "user2",
                    "acct_balance": 20,
                    "charges": [7, 12],
                    "meta": {"date": "11/4/2019", "lupdated": "now"},
                },
                {"meta": 2},
                2,
            ),
            (  # test integer greater
                [map_operations.map_clear("meta")],
                exp.GT(exp.IntBin("account_id"), 2),
                {"account_id": 3, "user_name": "user3", "acct_balance": 30, "charges": [8, 13], "meta": {}},
                {"meta": None},
                3,
            ),
            (  # test integer greatereq
                [map_operations.map_clear("meta")],
                exp.GE(exp.IntBin("account_id"), 2),
                {"account_id": 3, "user_name": "user3", "acct_balance": 30, "charges": [8, 13], "meta": {}},
                {"meta": None},
                3,
            ),
            (  # test integer less
                [list_operations.list_clear("charges")],
                exp.LT(exp.IntBin("account_id"), 5),
                {
                    "account_id": 4,
                    "user_name": "user4",
                    "acct_balance": 40,
                    "charges": [],
                    "meta": {"date": "11/4/2019"},
                },
                {},
                4,
            ),
            (  # test integer lesseq
                [list_operations.list_clear("charges")],
                exp.LE(exp.IntBin("account_id"), 4),
                {
                    "account_id": 4,
                    "user_name": "user4",
                    "acct_balance": 40,
                    "charges": [],
                    "meta": {"date": "11/4/2019"},
                },
                {},
                4,
            ),
            (  # test string unequal
                [list_operations.list_append("charges", 2)],
                exp.NE(exp.StrBin("user_name"), "user2"),
                {
                    "account_id": 4,
                    "user_name": "user4",
                    "acct_balance": 40,
                    "charges": [9, 14, 2],
                    "meta": {"date": "11/4/2019"},
                },
                {"charges": 3},
                4,
            ),
            (  # test not
                [list_operations.list_append("charges", 2)],
                exp.Not(exp.NE(exp.StrBin("user_name"), "user4")),
                {
                    "account_id": 4,
                    "user_name": "user4",
                    "acct_balance": 40,
                    "charges": [9, 14, 2],
                    "meta": {"date": "11/4/2019"},
                },
                {"charges": 3},
                4,
            ),
            (  # test string regex
                [list_operations.list_append("charges", 2)],
                exp.CmpRegex(aerospike.REGEX_ICASE, ".*4.*", exp.StrBin("user_name")),
                {
                    "account_id": 4,
                    "user_name": "user4",
                    "acct_balance": 40,
                    "charges": [9, 14, 2],
                    "meta": {"date": "11/4/2019"},
                },
                {"charges": 3},
                4,
            ),
            (  # test list or int
                [list_operations.list_append("charges", 2)],
                exp.Eq(exp.ListGetByValue(None, aerospike.LIST_RETURN_COUNT, 14, "charges"), 1),
                {
                    "account_id": 4,
                    "user_name": "user4",
                    "acct_balance": 40,
                    "charges": [9, 14, 2],
                    "meta": {"date": "11/4/2019"},
                },
                {"charges": 3},
                4,
            ),
            (  # test list and int
                [list_operations.list_append("charges", 2)],
                exp.LT(
                    exp.ListGetByRank(None, aerospike.LIST_RETURN_VALUE, exp.ResultType.INTEGER, -1, "charges"), 120
                ),
                {
                    "account_id": 4,
                    "user_name": "user4",
                    "acct_balance": 40,
                    "charges": [9, 14, 2],
                    "meta": {"date": "11/4/2019"},
                },
                {"charges": 3},
                4,
            ),
            (  # test list or str
                [list_operations.list_append("string_list", "s5")],
                exp.Eq(exp.ListGetByValue(None, aerospike.LIST_RETURN_COUNT, "s2", "string_list"), 1),
                {"string_list": ["s1", "s2", "s3", "s4", "s5"]},
                {"string_list": 5},
                5,
            ),
            # (# test list and str TODO Not sure how to write this in expressions
            #     [
            #         list_operations.list_remove_by_index_range('string_list', 0, aerospike.LIST_RETURN_VALUE, 2)
            #     ],
            #     exp.LT(
            #         exp.ListGetByRank(None, aerospike.LIST_RETURN_VALUE, exp.ResultType.INTEGER, -1, 'charges'),
            #         120
            #     ),
            #     exp.CmpRegex(aerospike.REGEX_ICASE, '.*s.*', exp.StrBin('user_name')),
            #     [
            #         exp.string_var('list_val'),
            #         exp.string_value('.*s.*'),
            #         exp.string_regex(aerospike.REGEX_ICASE),
            #         exp.list_bin('string_list'),
            #         exp.list_iterate_and('list_val')
            #     ],
            #     {
            #         'string_list': ['s3', 's4']
            #     },
            #     {'string_list': ['s1', 's2']},
            #     5
            # ),
            (  # test map_key_iterate_or
                [map_operations.map_put("map_bin", "k5", 5)],
                exp.Eq(exp.MapGetByKey(None, aerospike.MAP_RETURN_COUNT, exp.ResultType.INTEGER, "k3", "map_bin"), 1),
                {"map_bin": {"k1": 1, "k2": 2, "k3": 3, "k4": 4, "k5": 5}},
                {"map_bin": 5},
                6,
            ),
            (  # test map_key_iterate_and
                [map_operations.map_put("map_bin", "k5", 5)],
                exp.Eq(exp.MapGetByKey(None, aerospike.MAP_RETURN_COUNT, exp.ResultType.INTEGER, "k7", "map_bin"), 0),
                {"map_bin": {"k1": 1, "k2": 2, "k3": 3, "k4": 4, "k5": 5}},
                {"map_bin": 5},
                6,
            ),
            (  # test mapval_iterate_and
                [map_operations.map_put("map_bin", "k5", 5)],
                exp.Eq(exp.MapGetByValue(None, aerospike.MAP_RETURN_COUNT, 7, "map_bin"), 0),
                {"map_bin": {"k1": 1, "k2": 2, "k3": 3, "k4": 4, "k5": 5}},
                {"map_bin": 5},
                6,
            ),
            (  # test mapval_iterate_or
                [map_operations.map_get_by_key("map_bin", "k1", aerospike.MAP_RETURN_VALUE)],
                exp.Eq(exp.MapGetByValue(None, aerospike.MAP_RETURN_COUNT, 3, "map_bin"), 1),
                {"map_bin": {"k1": 1, "k2": 2, "k3": 3, "k4": 4}},
                {"map_bin": 1},
                6,
            ),
        ],
    )
    def test_expressions_key_operate(self, ops, expressions, expected_bins, expected_res, key_num):
        """
        Invoke the C client aerospike_key_operate with expressions.
        """
        key = ("test", "pred_evry", key_num)

        _, _, res = self.as_connection.operate(key, ops, policy={"expressions": expressions.compile()})
        assert res == expected_res

        _, _, bins = self.as_connection.get(key)
        assert bins == expected_bins

    @pytest.mark.parametrize(
        "ops, expressions, expected_bins, expected_res, key_num",
        [
            (  # test mapval_iterate_or
                [
                    map_operations.map_put_items("map_bin", {"k5": 5, "k6": 6}),
                    map_operations.map_get_by_key("map_bin", "k1", aerospike.MAP_RETURN_VALUE),
                ],
                exp.LT(exp.MapGetByRank(None, aerospike.MAP_RETURN_VALUE, exp.ResultType.INTEGER, 0, "map_bin"), 3),
                {"map_bin": {"k1": 1, "k2": 2, "k3": 3, "k4": 4, "k5": 5, "k6": 6}},
                [("map_bin", 6), ("map_bin", 1)],
                6,
            )
        ],
    )
    def test_expressions_key_operate_ordered(self, ops, expressions, expected_bins, expected_res, key_num):
        """
        Invoke the C client aerospike_key_operate with expressions using operate_ordered.
        """
        key = ("test", "pred_evry", key_num)

        _, _, res = self.as_connection.operate_ordered(key, ops, policy={"expressions": expressions.compile()})
        assert res == expected_res

        _, _, bins = self.as_connection.get(key)
        assert bins == expected_bins

    @pytest.mark.parametrize(
        "ops, expressions, expected, key_num",
        [
            (  # test mapval_iterate_or
                [map_operations.map_get_by_key("map_bin", "k1", aerospike.MAP_RETURN_VALUE)],
                exp.Not(
                    exp.Eq(exp.MapGetByRank(None, aerospike.MAP_RETURN_VALUE, exp.ResultType.INTEGER, 0, "map_bin"), 1)
                ),
                e.FilteredOut,
                6,
            )
        ],
    )
    def test_expressions_key_operate_ordered_negative(self, ops, expressions, expected, key_num):
        """
        Invoke the C client aerospike_key_operate with expressions using operate_ordered with expected failures.
        """
        key = ("test", "pred_evry", key_num)

        with pytest.raises(expected):
            _, _, res = self.as_connection.operate_ordered(key, ops, policy={"expressions": expressions.compile()})

    @pytest.mark.parametrize(
        "ops, expressions, key_num, bin",
        [
            (  # test geojson_within
                [operations.increment("id", 1)],
                exp.CmpGeo(exp.GeoBin("point"), geo_circle),
                7,
                "point",
            ),
            (  # test geojson_contains
                [operations.increment("id", 1)],
                exp.CmpGeo(exp.GeoBin("region"), geo_point),
                7,
                "point",
            ),
        ],
    )
    def test_expressions_key_operate_geojson(self, ops, expressions, key_num, bin):
        """
        Invoke the C client aerospike_key_operate with expressions.
        """
        key = ("test", "pred_evry", key_num)

        _, _, _ = self.as_connection.operate(key, ops, policy={"expressions": expressions.compile()})

        _, _, bins = self.as_connection.get(key)
        assert bins["id"] == 2

    # NOTE: may fail due to clock skew
    def test_expressions_key_operate_record_last_updated(self):
        """
        Invoke the C client aerospike_key_operate with a record_last_updated expressions.
        """

        for i in range(5):
            key = "test", "pred_lut", i
            self.as_connection.put(key, {"time": "earlier"})

        cutoff_nanos = (10**9) * int(time.time() + 2)
        time.sleep(5)

        for i in range(5, 10):
            key = "test", "pred_lut", i
            self.as_connection.put(key, {"time": "later"})

        results = []

        expr = exp.LT(exp.LastUpdateTime(), cutoff_nanos)
        ops = [operations.read("time")]

        for i in range(10):
            try:
                key = "test", "pred_lut", i
                _, _, res = self.as_connection.operate(key, ops, policy={"expressions": expr.compile()})
                results.append(res)
            except Exception:
                pass
            self.as_connection.remove(key)

        assert len(results) == 5
        for res in results:
            assert res["time"] == "earlier"

    # NOTE: may fail due to clock skew
    def test_expressions_key_operate_record_void_time(self):
        """
        Invoke the C client aerospike_key_operate with a rec_void_time expressions.
        """

        for i in range(5):
            key = "test", "pred_ttl", i
            self.as_connection.put(key, {"time": "earlier"}, meta={"ttl": 100})

        # 150 second range for record TTLs should be enough, we are storing with
        # Current time + 100s and current time +5000s, so only one of the group should be found
        void_time_range_start = (10**9) * int(time.time() + 50)
        void_time_range_end = (10**9) * int(time.time() + 150)

        for i in range(5, 10):
            key = "test", "pred_ttl", i
            self.as_connection.put(key, {"time": "later"}, meta={"ttl": 1000})

        results = []

        expr = exp.And(
            exp.GT(exp.VoidTime(), void_time_range_start), exp.LT(exp.VoidTime(), void_time_range_end)
        ).compile()

        ops = [operations.read("time")]

        for i in range(10):
            try:
                key = "test", "pred_ttl", i
                _, _, res = self.as_connection.operate(key, ops, policy={"expressions": expr})
                results.append(res)
            except Exception:
                pass
            self.as_connection.remove(key)

        assert len(results) == 5
        for res in results:
            assert res["time"] == "earlier"

    def test_expressions_key_operate_record_digest_modulo(self):
        """
        Invoke the C client aerospike_key_operate with a rec_digest_modulo expressions.
        """

        less_than_128 = 0
        for i in range(100):
            key = "test", "demo", i
            if aerospike.calc_digest(*key)[-1] < 128:
                less_than_128 += 1
            self.as_connection.put(key, {"dig_id": i})

        results = []

        expr = exp.LT(exp.DigestMod(256), 128).compile()

        ops = [operations.read("dig_id")]

        for i in range(100):
            try:
                key = "test", "demo", i
                _, _, res = self.as_connection.operate(key, ops, policy={"expressions": expr})
                results.append(res)
            except Exception:
                pass
            self.as_connection.remove(key)

        assert len(results) == less_than_128

    @pytest.mark.parametrize(
        "ops, expressions, expected, key_num",
        [
            (  # filtered out
                [operations.increment("account_id", 1)],
                exp.Eq(exp.IntBin("account_id"), 5).compile(),
                e.FilteredOut,
                3,
            ),
            (  # incorrect bin type
                [
                    list_operations.list_remove_by_index_range("charges", 0, 3, aerospike.LIST_RETURN_COUNT),
                    operations.increment("acct_balance", -23),
                ],
                exp.And(exp.GE(exp.StrBin("acct_balance"), 10), exp.LE(exp.IntBin("acct_balance"), 50)).compile(),
                e.InvalidRequest,
                4,
            ),
            (  # filtered out
                [map_operations.map_put("meta", "lupdated", "now")],
                exp.Not(
                    exp.Or(exp.Eq(exp.StrBin("user_name"), "user2"), exp.GE(exp.IntBin("acct_balance"), 50))
                ).compile(),
                e.FilteredOut,
                2,
            ),
            ([map_operations.map_put("meta", "lupdated", "now")], [], e.ParamError, 2),  # empty expressions list
            (  # expressions not in list
                [map_operations.map_put("meta", "lupdated", "now")],
                ("bad expressions",),
                e.ParamError,
                2,
            ),
        ],
    )
    def test_expressions_key_operate_negative(self, ops, expressions, expected, key_num):
        """
        Invoke the C client aerospike_key_operate with expressions. Expecting failures.
        """
        key = ("test", "pred_evry", key_num)

        with pytest.raises(expected):
            self.as_connection.operate(key, ops, policy={"expressions": expressions})

    @pytest.mark.parametrize(
        "expressions, rec_place, rec_bin, expected",
        [
            (exp.Eq(exp.IntBin("account_id"), 2), 1, "account_id", 2),
            (exp.Eq(exp.StrBin("user_name"), "user2"), 1, "account_id", 2),
            (
                exp.Or(exp.Eq(exp.StrBin("user_name"), "user2"), exp.GE(exp.IntBin("acct_balance"), 30)),
                2,
                "account_id",
                3,
            ),
        ],
    )
    def test_pos_get_many_with_expressions(self, expressions, rec_place, rec_bin, expected):
        """
        Proper call to get_many with expressions in policy
        """
        records = self.as_connection.get_many(self.keys, {"expressions": expressions.compile()})

        # assert isinstance(records, list)
        # assert records[2][2]['age'] == 2
        assert records[rec_place][2][rec_bin] == expected

    def test_pos_get_many_with_large_expressions(self):
        """
        Proper call to get_many with expressions in policy.
        """
        expr = exp.Or(
            exp.Eq(exp.IntBin("account_id"), 4),
            exp.Eq(exp.StrBin("user_name"), "user3"),
            exp.LT(exp.ListGetByRank(None, aerospike.LIST_RETURN_VALUE, exp.ResultType.INTEGER, -1, "charges"), 12),
        )

        matched_recs = []
        records = self.as_connection.get_many(self.keys, {"expressions": expr.compile()})
        for rec in records:
            if rec[2] is not None:
                matched_recs.append(rec[2])

        assert len(matched_recs) == 3
        for rec in matched_recs:
            assert rec["account_id"] == 1 or rec["account_id"] == 3 or rec["account_id"] == 4

    def test_pos_select_many_with_large_expressions(self):
        """
        Proper call to select_many with expressions in policy.
        """
        expr = exp.Or(
            exp.Eq(exp.IntBin("account_id"), 4),
            exp.Eq(exp.StrBin("user_name"), "user3"),
            exp.LT(exp.ListGetByRank(None, aerospike.LIST_RETURN_VALUE, exp.ResultType.INTEGER, -1, "charges"), 12),
        )

        matched_recs = []
        records = self.as_connection.select_many(self.keys, ["account_id"], {"expressions": expr.compile()})
        for rec in records:
            if rec[2] is not None:
                matched_recs.append(rec[2])

        assert len(matched_recs) == 3
        for rec in matched_recs:
            assert rec["account_id"] == 1 or rec["account_id"] == 3 or rec["account_id"] == 4

    def test_pos_remove_with_expressions(self):
        """
        Call remove with expressions in policy.
        """
        expr = exp.Eq(exp.IntBin("account_id"), 1)
        self.as_connection.remove(self.keys[0], {"expressions": expr.compile()})

        rec = self.as_connection.exists(self.keys[0])
        assert rec[1] is None

    def test_remove_with_expressions_filtered_out(self):
        """
        Call remove with expressions in policy with expected failure.
        """
        expr = exp.Eq(exp.IntBin("account_id"), 3)
        with pytest.raises(e.FilteredOut):
            self.as_connection.remove(self.keys[0], policy={"expressions": expr.compile()})

    def test_remove_bin_with_expressions(self):
        """
        Call remove_bin with expressions in policy.
        """
        expr = exp.Eq(exp.IntBin("account_id"), 1)
        self.as_connection.remove_bin(self.keys[0], ["account_id", "user_name"], policy={"expressions": expr.compile()})

        rec = self.as_connection.get(self.keys[0])
        assert rec[2].get("account_id") is None and rec[2].get("user_name") is None

    def test_remove_bin_with_expressions_filtered_out(self):
        """
        Call remove_bin with expressions in policy with expected failure.
        """
        expr = exp.Eq(exp.IntBin("account_id"), 4)
        with pytest.raises(e.FilteredOut):
            self.as_connection.remove_bin(
                self.keys[0], ["account_id", "user_name"], policy={"expressions": expr.compile()}
            )

    def test_put_with_expressions(self):
        """
        Call put with expressions in policy.
        """
        expr = exp.Eq(exp.IntBin("account_id"), 1)
        self.as_connection.put(self.keys[0], {"newkey": "newval"}, policy={"expressions": expr.compile()})

        rec = self.as_connection.get(self.keys[0])
        assert rec[2]["newkey"] == "newval"

    def test_put_new_record_with_expressions(self):  # should this fail?
        """
        Call put a new record with expressions in policy.
        """
        expr = exp.Eq(exp.IntBin("account_id"), 1)
        key = ("test", "demo", 10)
        self.as_connection.put(key, {"newkey": "newval"}, policy={"expressions": expr.compile()})

        rec = self.as_connection.get(key)
        self.as_connection.remove(key)
        assert rec[2]["newkey"] == "newval"

    def test_put_with_expressions_filtered_out(self):
        """
        Call put with expressions in policy with expected failure.
        """
        expr = exp.Eq(exp.IntBin("account_id"), 4)
        with pytest.raises(e.FilteredOut):
            self.as_connection.put(self.keys[0], {"newkey": "newval"}, policy={"expressions": expr.compile()})

    def test_get_with_expressions(self):
        """
        Call to get with expressions in policy.
        """
        expr = exp.Eq(exp.IntBin("account_id"), 1)
        record = self.as_connection.get(self.keys[0], {"expressions": expr.compile()})

        assert record[2]["account_id"] == 1

    def test_get_with_expressions_filtered_out(self):
        """
        Call to get with expressions in policy with expected failures.
        """
        expr = exp.Eq(exp.IntBin("account_id"), 3)
        with pytest.raises(e.FilteredOut):
            self.as_connection.get(self.keys[0], {"expressions": expr.compile()})

    def test_select_with_expressions(self):
        """
        Call to select with expressions in policy.
        """

        expr = exp.And(
            exp.Eq(exp.IntBin("acct_balance"), 20),
            exp.LT(exp.ListGetByRank(None, aerospike.LIST_RETURN_VALUE, exp.ResultType.INTEGER, -1, "charges"), 20),
        )

        result = self.as_connection.select(
            self.keys[1], ["account_id", "acct_balance"], {"expressions": expr.compile()}
        )
        assert result[2]["account_id"] == 2 and result[2]["acct_balance"] == 20

    def test_select_with_expressions_filtered_out(self):
        """
        Call to select with expressions in policy with expected failures.
        """
        expr = exp.And(
            exp.Eq(exp.IntBin("acct_balance"), 20),
            exp.LT(exp.ListGetByRank(None, aerospike.LIST_RETURN_VALUE, exp.ResultType.INTEGER, -1, "charges"), 10),
        )

        with pytest.raises(e.FilteredOut):
            self.as_connection.select(self.keys[1], ["account_id", "acct_balance"], {"expressions": expr.compile()})

    def test_exists_many_with_large_expressions(self):
        """
        Proper call to exists_many with expressions in policy.
        """

        expr = exp.Or(
            exp.Eq(exp.IntBin("account_id"), 4),
            exp.Eq(exp.StrBin("user_name"), "user3"),
            exp.LT(exp.ListGetByRank(None, aerospike.LIST_RETURN_VALUE, exp.ResultType.INTEGER, -1, "charges"), 12),
        )

        matched_recs = []
        records = self.as_connection.exists_many(self.keys, {"expressions": expr.compile()})
        for rec in records:
            if rec[1] is not None:
                matched_recs.append(rec[1])

        assert len(matched_recs) == 3



# -*- coding: utf-8 -*-
import pytest
import time

import aerospike
from aerospike import exception, predicates
from aerospike_helpers.operations import operations, map_operations
from aerospike_helpers import expressions as exp

TEST_NS = "test"
TEST_SET = "background_scan1"
TEST_SET2 = "background_scan2"
TEST_SET3 = "background_scan3"
TEST_UDF_MODULE = "bin_lua"
TEST_UDF_FUNCTION = "mytransformscan"
# Hack to get long to exist in Python 3
try:
    long
except NameError:
    long = int


def wait_for_job_completion(as_connection, job_id):
    """
    Blocks until the job has completed
    """
    time.sleep(0.1)
    while True:
        response = as_connection.job_info(job_id, aerospike.JOB_SCAN)
        if response["status"] != aerospike.JOB_STATUS_INPROGRESS:
            break
        time.sleep(0.1)


def add_test_udf(client):
    policy = {}
    client.udf_put("bin_lua.lua", 0, policy)


def drop_test_udf(client):
    client.udf_remove("bin_lua.lua")


def validate_records(client, keys, validator):
    for key in keys:
        _, _, rec = client.get(key)
        assert validator(rec)


class TestScanApply(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, connection_with_config_funcs):
        add_test_udf(self.as_connection)
        keys = [(TEST_NS, TEST_SET, i) for i in range(50)]
        for i, key in enumerate(keys):
            self.as_connection.put(key, {"number": i})
        keys = [(TEST_NS, TEST_SET2, i) for i in range(10)]
        for i, key in enumerate(keys):
            self.as_connection.put(key, {"number": i})
        keys = [(TEST_NS, TEST_SET3, i) for i in range(10)]
        for i, key in enumerate(keys):
            self.as_connection.put(key, {"numbers": {1: i, 2: 2 * i, 3: 3 * i}})

        def teardown():
            drop_test_udf(self.as_connection)
            keys = [(TEST_NS, TEST_SET, i) for i in range(50)]
            for i, key in enumerate(keys):
                self.as_connection.remove(key, {"number": i})
            keys = [(TEST_NS, TEST_SET2, i) for i in range(10)]
            for i, key in enumerate(keys):
                self.as_connection.remove(key, {"number": i})

        request.addfinalizer(teardown)

    def test_background_execute_return_val(self):
        """
        Ensure that Scan.execute_background() returns an int like object.
        """
        test_bin = "St1"
        scan = self.as_connection.scan(TEST_NS, TEST_SET)
        scan.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        res = scan.execute_background()
        assert isinstance(res, (int, long))

    def test_background_execute_no_predicate(self):
        """
        Ensure that Scan.execute_background() gets applied to all records.
        """
        test_bin = "number"
        keys = [(TEST_NS, TEST_SET, i) for i in range(50)]

        scan = self.as_connection.scan(TEST_NS, TEST_SET)
        scan.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin, 1])
        job_id = scan.execute_background()
        wait_for_job_completion(self.as_connection, job_id)

        for i, key in enumerate(keys):
            _, _, bins = self.as_connection.get(key)
            assert bins["number"] == i + 1

    def test_background_execute_expressions_everywhere(self):
        """
        Ensure that Scan.execute_background() gets applied to records that match the expressions.
        """
        test_bin = "number"
        keys = [(TEST_NS, TEST_SET, i) for i in range(50)]

        expr = exp.Or(exp.Eq(exp.IntBin("number"), 2), exp.Eq(exp.IntBin("number"), 3))

        policy = {"expressions": expr.compile()}

        scan = self.as_connection.scan(TEST_NS, TEST_SET)
        scan.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin, 1])
        job_id = scan.execute_background(policy)
        wait_for_job_completion(self.as_connection, job_id)

        for i, key in enumerate(keys):
            _, _, bins = self.as_connection.get(key)
            if i == 2 or i == 3:
                assert bins[test_bin] == i + 1
            else:
                assert bins.get(test_bin) == i

    @pytest.mark.xfail(reason="scan does not implement .where()")
    def test_background_execute_expressions_and_predicate(self):
        """
        Ensure that Scan.execute_background() gets applied to records that match the predicate.
        NOTE: the predicate overrides the expressions
        """
        test_bin = "Stpredold"
        keys = [(TEST_NS, TEST_SET, i) for i in range(50)]

        expr = exp.Or(exp.Eq(exp.IntBin("number"), 2), exp.Eq(exp.IntBin("number"), 3))

        policy = {"expressions": expr.compile()}

        number_predicate = predicates.equals("number", 4)

        scan = self.as_connection.scan(TEST_NS, TEST_SET)
        scan.where(number_predicate)
        scan.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        job_id = scan.execute_background(policy)
        wait_for_job_completion(self.as_connection, job_id)

        for key in keys:
            _, _, bins = self.as_connection.get(key)
            if bins["number"] == 4:
                assert bins[test_bin] == "aerospike"
            else:
                assert bins.get(test_bin) is None

    def test_background_execute_with_ops_and_expressions(self):
        """
        Ensure that Scan.execute_background() applies ops to records that match the expressions.
        """
        test_bin = "Stops_preds"
        keys = [(TEST_NS, TEST_SET, i) for i in range(50)]

        scan = self.as_connection.scan(TEST_NS, TEST_SET)
        # scan.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])

        ops = [operations.write(test_bin, "new_val")]

        expr = exp.Or(exp.Eq(exp.IntBin("number"), 2), exp.Eq(exp.IntBin("number"), 3))

        policy = {"expressions": expr.compile()}

        scan.add_ops(ops)
        job_id = scan.execute_background(policy)
        wait_for_job_completion(self.as_connection, job_id)

        for key in keys:
            _, _, bins = self.as_connection.get(key)
            if bins["number"] == 2 or bins["number"] == 3:
                assert bins[test_bin] == "new_val"
            else:
                assert bins.get(test_bin) is None

    def test_background_execute_with_ops_and_expressions_None_set(self):
        """
        Ensure that Scan.execute_background() applies ops to all records in the NS that match the expressions.
        """
        test_bin = "Stops_noset"
        keys = [(TEST_NS, TEST_SET, i) for i in range(50)]
        keys2 = [(TEST_NS, TEST_SET2, i) for i in range(10)]
        scan = self.as_connection.scan(TEST_NS, None)
        # scan.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])

        ops = [operations.write(test_bin, "new_val")]

        expr = exp.Or(exp.Eq(exp.IntBin("number"), 2), exp.Eq(exp.IntBin("number"), 3))

        policy = {"expressions": expr.compile()}

        scan.add_ops(ops)
        job_id = scan.execute_background(policy)
        wait_for_job_completion(self.as_connection, job_id)

        for key in keys:
            _, _, bins = self.as_connection.get(key)
            if bins["number"] == 2 or bins["number"] == 3:
                assert bins[test_bin] == "new_val"
            else:
                assert bins.get(test_bin) is None

        for key in keys2:
            _, _, bins = self.as_connection.get(key)
            if bins["number"] == 2 or bins["number"] == 3:
                assert bins[test_bin] == "new_val"
            else:
                assert bins.get(test_bin) is None

    def test_background_execute_with_ops(self):
        """
        Ensure that Scan.execute_background() applies ops to all records.
        """
        test_bin = "Stops"
        keys = [(TEST_NS, TEST_SET, i) for i in range(50)]

        scan = self.as_connection.scan(TEST_NS, TEST_SET)
        # scan.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])

        ops = [operations.write(test_bin, "new_val")]

        scan.add_ops(ops)
        job_id = scan.execute_background()
        wait_for_job_completion(self.as_connection, job_id)

        validate_records(self.as_connection, keys, lambda rec: rec[test_bin] == "new_val")

    def test_background_execute_with_map_ops(self):
        """
        Ensure that Scan.execute_background() applies ops to matching map bins.
        """
        test_bin = "numbers"
        keys = [(TEST_NS, TEST_SET3, i) for i in range(10)]

        scan = self.as_connection.scan(TEST_NS, TEST_SET3)

        ops = [map_operations.map_remove_by_value_range(test_bin, 20, 35, aerospike.MAP_RETURN_NONE)]

        scan.add_ops(ops)
        job_id = scan.execute_background()
        wait_for_job_completion(self.as_connection, job_id)

        records = self.as_connection.get_many(keys)
        for i, rec in enumerate(records):
            if i * 3 < 20:
                assert rec[2]["numbers"] == {1: i, 2: i * 2, 3: i * 3}
            else:
                assert rec[2]["numbers"] == {1: i, 2: i * 2}

    @pytest.mark.xfail(reason="Scan does not implement .where()")
    def test_background_execute_with_ops_and_preds(self):
        """
        Ensure that Scan.execute_background() applies ops to records that match the predicate.
        """
        test_bin = "St1"
        keys = [(TEST_NS, TEST_SET, i) for i in range(50)]

        scan = self.as_connection.scan(TEST_NS, TEST_SET)
        number_predicate = predicates.equals("number", 3)
        # scan.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])

        ops = [operations.append(test_bin, "new_val")]

        scan.add_ops(ops)
        scan.where(number_predicate)
        job_id = scan.execute_background()
        wait_for_job_completion(self.as_connection, job_id)

        _, _, num_5_record = self.as_connection.get((TEST_NS, TEST_SET, 5))
        assert num_5_record[test_bin] == "aerospike"

        _, _, num_3_record = self.as_connection.get((TEST_NS, TEST_SET, 3))
        assert num_3_record[test_bin] == "aerospikenew_val"

        # cleanup
        ops = [operations.write(test_bin, "aerospike")]

        scan.add_ops(ops)
        job_id = scan.execute_background()
        wait_for_job_completion(self.as_connection, job_id)

        validate_records(self.as_connection, keys, lambda rec: rec[test_bin] == "aerospike")

    @pytest.mark.xfail(reason="Scan does not implement .where()")
    def test_background_execute_sindex_predicate(self):
        """
        Ensure that Scan.execute_background() only applies to records matched by
        the specified predicate.
        """
        test_bin = "St3"
        keys = [(TEST_NS, TEST_SET, i) for i in range(50)]
        number_predicate = predicates.equals("number", 5)

        scan = self.as_connection.scan(TEST_NS, TEST_SET)
        scan.where(number_predicate)
        scan.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        job_id = scan.execute_background()
        wait_for_job_completion(self.as_connection, job_id)

        keys = [(TEST_NS, TEST_SET, i) for i in range(50) if i != 5]
        validate_records(self.as_connection, keys, lambda rec: test_bin not in rec)
        _, _, num_5_record = self.as_connection.get((TEST_NS, TEST_SET, 5))
        assert num_5_record[test_bin] == "aerospike"

    def test_background_execute_with_policy(self):
        """
        Ensure that Scan.execute_background() returns an int like object.
        """
        test_bin = "St5"
        scan = self.as_connection.scan(TEST_NS, TEST_SET)
        scan.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        res = scan.execute_background({})
        assert isinstance(res, (int, long))

    def test_background_execute_with_policy_kwarg(self):
        """
        Ensure that Scan.execute_background() returns an int like object.
        """
        test_bin = "St6"
        scan = self.as_connection.scan(TEST_NS, TEST_SET)
        scan.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        res = scan.execute_background(policy={})
        assert isinstance(res, (int, long))

    def test_background_execute_with_invalid_policy_type(self):
        """
        Ensure that Scan.execute_background() returns an int like object.
        """
        test_bin = "St6"
        scan = self.as_connection.scan(TEST_NS, TEST_SET)
        scan.apply(TEST_UDF_MODULE, TEST_UDF_FUNCTION, [test_bin])
        # Policy needs to be a dict. Not a string
        with pytest.raises(exception.ParamError):
            scan.execute_background("Honesty is the best Policy")



# -*- coding: utf-8 -*-
import pytest
import random
from aerospike import exception as e

import aerospike


class TestListSet(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        keys = []
        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "contact_no": [i, i + 1], "city": ["Pune", "Dehli"]}
            self.as_connection.put(key, rec)
            keys.append(key)
        key = ("test", "demo", 2)
        self.as_connection.list_append(key, "contact_no", [45, 50, 80])
        keys.append(key)

        def teardown():
            """
            Teardown method.
            """
            for key in keys:
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "key, field, index, value, expected",
        [
            (
                ("test", "demo", 1),  # list element with integer
                "contact_no",
                5,
                1000,
                {"city": ["Pune", "Dehli"], "contact_no": [1, 2, None, None, None, 1000], "name": "name1"},
            ),
            (
                ("test", "demo", 1),  # list element with list
                "contact_no",
                5,
                [500, 1000],
                {"city": ["Pune", "Dehli"], "contact_no": [1, 2, None, None, None, [500, 1000]], "name": "name1"},
            ),
            (
                ("test", "demo", 1),  # list element with string
                "contact_no",
                5,
                "string",
                {"city": ["Pune", "Dehli"], "contact_no": [1, 2, None, None, None, "string"], "name": "name1"},
            ),
            (
                ("test", "demo", 1),  # float
                "contact_no",
                5,
                45.896,
                {"city": ["Pune", "Dehli"], "contact_no": [1, 2, None, None, None, 45.896], "name": "name1"},
            ),
            (
                ("test", "demo", 1),  # Boolean
                "contact_no",
                5,
                False,
                {"city": ["Pune", "Dehli"], "contact_no": [1, 2, None, None, None, 0], "name": "name1"},
            ),
            (
                ("test", "demo", 1),  # Bytearray
                "contact_no",
                0,
                bytearray("asd;as[d'as;d", "utf-8"),
                {"contact_no": [bytearray(b"asd;as[d'as;d"), 2], "city": ["Pune", "Dehli"], "name": "name1"},
            ),
        ],
    )
    def test_pos_list_set_with_elements(self, key, field, index, value, expected):
        """
        Invoke list_set() sets
        """
        status = self.as_connection.list_set(key, field, index, value)
        assert status == 0

        key, _, bins = self.as_connection.get(key)
        assert bins == expected

    def test_pos_list_set_with_element_map(self):
        """
        Invoke list_set() sets list element with map
        """
        key = ("test", "demo", 1)

        status = self.as_connection.list_set(key, "contact_no", 5, {"k1": 56})
        assert status == 0

        key, _, bins = self.as_connection.get(key)
        assert bins == {"city": ["Pune", "Dehli"], "contact_no": [1, 2, None, None, None, {"k1": 56}], "name": "name1"}

    def test_pos_list_set_with_correct_policy(self):
        """
        Invoke list_append() append list with correct policy
        """
        key = ("test", "demo", 2)
        policy = {
            "timeout": 180000,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
        }

        status = self.as_connection.list_set(key, "city", 7, "Mumbai", {}, policy)
        assert status == 0

        key, _, bins = self.as_connection.get(key)
        assert bins == {
            "city": ["Pune", "Dehli", None, None, None, None, None, "Mumbai"],
            "contact_no": [2, 3, [45, 50, 80]],
            "name": "name2",
        }

    def test_pos_list_set_with_nonexistent_key(self):
        """
        Invoke list_set() with non-existent key
        """
        charSet = "abcdefghijklmnopqrstuvwxyz1234567890"
        minLength = 5
        maxLength = 30
        length = random.randint(minLength, maxLength)
        key = ("test", "demo", "".join(map(lambda unused: random.choice(charSet), range(length))) + ".com")
        status = self.as_connection.list_set(key, "contact_no", 0, 100)
        assert status == 0
        key, _, bins = self.as_connection.get(key)
        assert bins == {"contact_no": [100]}
        self.as_connection.remove(key)

    # Negative Tests
    def test_neg_list_set_with_no_parameters(self):
        """
        Invoke list_set() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_set()
        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_neg_list_set_with_incorrect_policy(self):
        """
        Invoke list_set() with incorrect policy
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 0.5}
        try:
            self.as_connection.list_set(key, "contact_no", 0, 850, {}, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"

    def test_neg_list_set_with_nonexistent_bin(self):
        """
        Invoke list_set() with non-existent bin
        """
        key = ("test", "demo", 1)
        charSet = "abcdefghijklmnopqrstuvwxyz1234567890"
        minLength = 5
        maxLength = 10
        length = random.randint(minLength, maxLength)
        bin = "".join(map(lambda unused: random.choice(charSet), range(length))) + ".com"
        try:
            self.as_connection.list_set(key, bin, 0, 75)

        except e.BinIncompatibleType as exception:
            assert exception.code == 12

    def test_neg_list_set_with_extra_parameter(self):
        """
        Invoke list_set() with extra parameter.
        """
        key = ("test", "demo", 1)
        policy = {}
        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_set(key, "contact_no", 1, 999, {}, policy, "")

        assert "list_set() takes at most 6 arguments (7 given)" in str(typeError.value)

    def test_neg_list_set_policy_is_string(self):
        """
        Invoke list_set() with policy is string
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_set(key, "contact_no", 1, 30, {}, "")

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "policy must be a dict"

    def test_neg_list_set_key_is_none(self):
        """
        Invoke list_set() with key is none
        """
        try:
            self.as_connection.list_set(None, "contact_no", 0, 89)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "key is invalid"

    def test_neg_list_set_bin_is_none(self):
        """
        Invoke list_set() with bin is none
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_set(key, None, 1, 555)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Bin name should be of type string"

    def test_neg_list_set_with_negative_index(self):
        """
        Invoke list_set() with negative index
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_set(key, "contact_no", -56, 12)
        except e.OpNotApplicable as exception:
            assert exception.code == 26

    def test_neg_list_set_meta_type_integer(self):
        """
        Invoke list_set() with metadata input is of type integer
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_set(key, "contact_no", 0, 679, 888)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Metadata should be of type dictionary"

    def test_neg_list_set_index_type_string(self):
        """
        Invoke list_set() with index is of type string
        """
        key = ("test", "demo", 1)

        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_set(key, "contact_no", "Fifth", 448)
        assert "an integer is required" or "cannot be interpreted as an integer" in str(typeError.value)



# -*- coding: utf-8 -*-
import pytest
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


@pytest.mark.xfail(TestBaseClass.tls_in_use(), reason="get_nodes may fail when using TLS")
@pytest.mark.usefixtures("as_connection")
class TestGetNodes(object):
    """
    Test Cases for the use of aerospike.Client.get_nodes
    method
    """

    def test_pos_get_nodes(self):
        """
        Test successful to aerospike.Client.get_nodes
        without parameters.

        It should return something like [('1.1.1.1', 3000)]
        """
        response = self.as_connection.get_nodes()
        assert response is not None
        assert isinstance(response, list)
        assert isinstance(response[0], tuple)
        assert len(response[0]) == 2

    def test_get_nodes_with_parameter(self):
        """
        Test that a call to aerospike.Client.get_nodes with
        a string parameter does not raise an error
        """
        response = self.as_connection.get_nodes("parameter")
        assert response is not None

    # Tests for behaviors that raise errors
    def test_pos_get_nodes_without_connection(self):
        """
        Test that an attempt to call get_nodes before a connection
        is established will raise the expected error
        """
        config = TestBaseClass.get_connection_config()
        unconnected_client = aerospike.client(config)
        unconnected_client.close()

        try:
            unconnected_client.get_nodes()

        except e.ClusterError as exception:
            assert exception.code == 11
            assert exception.msg == "No connection to aerospike cluster"



# -*- coding: utf-8 -*-
import pytest
from aerospike import exception as e
from aerospike_helpers.operations import map_operations as mop
from aerospike_helpers.operations import list_operations as lop
from aerospike import KeyOrderedDict
import aerospike_helpers.cdt_ctx as ctx

import aerospike


class TestOrderedDictGetByValue(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        map_policy = {"map_order": aerospike.MAP_KEY_VALUE_ORDERED}

        self.list_key = ("test", "demo", 100)
        self.as_connection.put(self.list_key, {"map_list": []})

        self.map_key = ("test", "demo", 200)
        self.as_connection.put(self.map_key, {"map_of_maps": {}})

        lmap_ctx1 = ctx.cdt_ctx_list_index(0)
        lmap_ctx2 = ctx.cdt_ctx_list_index(1)
        lmap_ctx3 = ctx.cdt_ctx_list_index(2)

        mmap_ctx1 = ctx.cdt_ctx_map_key("key1")
        mmap_ctx2 = ctx.cdt_ctx_map_key("key2")
        mmap_ctx3 = ctx.cdt_ctx_map_key("key3")

        my_dict1 = {"a": 1, "b": 2, "c": 3}
        my_dict2 = {"d": 4, "e": 5, "f": 6}
        my_dict3 = {"g": 7, "h": 8, "i": 9}

        ops = [
            lop.list_append_items("map_list", [my_dict1, my_dict2, my_dict3]),
            mop.map_set_policy("map_list", map_policy, [lmap_ctx1]),
            mop.map_set_policy("map_list", map_policy, [lmap_ctx2]),
            mop.map_set_policy("map_list", map_policy, [lmap_ctx3]),
        ]
        self.as_connection.operate(self.list_key, ops)

        ops = [
            mop.map_put_items(
                "map_of_maps", {"key1": my_dict1, "key2": my_dict2, "key3": my_dict2}, map_policy=map_policy
            ),
            mop.map_set_policy("map_of_maps", map_policy, [mmap_ctx1]),
            mop.map_set_policy("map_of_maps", map_policy, [mmap_ctx2]),
            mop.map_set_policy("map_of_maps", map_policy, [mmap_ctx3]),
        ]
        self.as_connection.operate(self.map_key, ops)

        yield

        try:
            self.as_connection.remove(self.list_key)
            self.as_connection.remove(self.map_key)
        except e.AerospikeError:
            pass

    def test_list_get_by_value_ordered_dict(self):
        """
        OrderedDicts get converted to key sorted as_maps, test if they match key ordered map bins on the server.
        """

        element = KeyOrderedDict({"f": 6, "e": 5, "d": 4})

        ops = [lop.list_get_by_value("map_list", element, aerospike.LIST_RETURN_COUNT)]
        _, _, res = self.as_connection.operate(self.list_key, ops)

        assert res == {"map_list": 1}

    def test_list_remove_by_value_ordered_dict(self):
        """
        OrderedDicts get converted to key sorted as_maps, test if they match key ordered map bins on the server.
        """

        element = KeyOrderedDict({"f": 6, "e": 5, "d": 4})

        ops = [lop.list_remove_by_value("map_list", element, aerospike.LIST_RETURN_COUNT)]
        _, _, res = self.as_connection.operate(self.list_key, ops)

        assert res == {"map_list": 1}

    def test_map_get_by_value_ordered_dict(self):
        """
        OrderedDicts get converted to key sorted as_maps, test if they match key ordered map bins on the server.
        """

        element = KeyOrderedDict({"f": 6, "e": 5, "d": 4})

        ops = [mop.map_get_by_value("map_of_maps", element, aerospike.MAP_RETURN_COUNT)]
        _, _, res = self.as_connection.operate(self.map_key, ops)

        assert res == {"map_of_maps": 2}



# These fail if we don't have a new server
# -*- coding: utf-8 -*-
import pytest
import aerospike
from aerospike import exception as e
from aerospike_helpers.operations import map_operations as map_ops


class TestMapWriteMode(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        self.keys = []

        yield

        for key in self.keys:
            self.as_connection.remove(key)

    def test_default_allows_update_and_create(self):
        key = "test", "write_mode", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {"map_write_mode": aerospike.MAP_UPDATE}
        ops = [
            map_ops.map_put("map", "existing", "new", map_policy=map_policy),
            map_ops.map_put("map", "new", "new", map_policy=map_policy),
        ]
        self.as_connection.operate(key, ops)

        _, _, bins = self.as_connection.get(key)

        map_bin = bins["map"]
        assert map_bin["existing"] == "new"
        assert map_bin["new"] == "new"

    def test_create_only_does_not_allow_update(self):
        key = "test", "write_mode", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {"map_write_mode": aerospike.MAP_CREATE_ONLY}
        ops = [
            map_ops.map_put("map", "existing", "new", map_policy=map_policy),
        ]
        with pytest.raises(e.ElementExistsError):
            self.as_connection.operate(key, ops)

        _, _, bins = self.as_connection.get(key)

        map_bin = bins["map"]
        assert map_bin["existing"] == "old"

    def test_create_only_allows_create(self):
        key = "test", "write_mode", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {"map_write_mode": aerospike.MAP_CREATE_ONLY}
        ops = [
            map_ops.map_put("map", "new", "new", map_policy=map_policy),
        ]
        self.as_connection.operate(key, ops)
        _, _, bins = self.as_connection.get(key)

        map_bin = bins["map"]
        assert map_bin["existing"] == "old"
        assert map_bin["new"] == "new"

    def test_update_only_does_not_allow_create(self):
        key = "test", "write_mode", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {"map_write_mode": aerospike.MAP_UPDATE_ONLY}
        ops = [
            map_ops.map_put("map", "new", "new", map_policy=map_policy),
        ]

        with pytest.raises(e.ElementNotFoundError):
            self.as_connection.operate(key, ops)

        _, _, bins = self.as_connection.get(key)

        map_bin = bins["map"]
        assert map_bin["existing"] == "old"
        assert "new" not in map_bin

    def test_update_only_allows_update(self):
        key = "test", "write_mode", 1
        self.keys.append(key)
        self.as_connection.put(key, {"map": {"existing": "old"}})

        map_policy = {"map_write_mode": aerospike.MAP_UPDATE_ONLY}
        ops = [
            map_ops.map_put("map", "existing", "new", map_policy=map_policy),
        ]

        self.as_connection.operate(key, ops)

        _, _, bins = self.as_connection.get(key)

        map_bin = bins["map"]
        assert map_bin["existing"] == "new"



# Insert records for 3 players and their scores
keyTuples = [("test", "demo", i) for i in range(1, 4)]
bins = [
    {"scores": [1, 4, 3, 10]},
    {"scores": [20, 1, 4, 28]},
    {"scores": [50, 20, 10, 20]},
]
for keyTuple, bin in zip(keyTuples, bins):
    client.put(keyTuple, bin)

# Get highest scores for each player
from aerospike_helpers.operations import list_operations
ops = [
    list_operations.list_get_by_rank("scores", -1, aerospike.LIST_RETURN_VALUE)
]
records = client.batch_get_ops(keyTuples, ops)

# Print results
for _, _, bins in records:
    print(bins)
# {'scores': 10}
# {'scores': 28}
# {'scores': 50}



# -*- coding: utf-8 -*-
from __future__ import print_function


from distutils.version import LooseVersion
import pytest
from .as_status_codes import AerospikeStatus
from .udf_helpers import wait_for_udf_removal, wait_for_udf_to_exist
from .test_base_class import TestBaseClass
import aerospike
from aerospike import exception as e


def is_greater_451(version_str):
    """
    Is the server version 4.5.1.0-pre or newer
    """
    return LooseVersion(version_str) >= LooseVersion("4.5.1")


class TestUdfRemove(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method, adds a UDF and waits for the UDF to exist
        """
        as_connection.udf_put(self.udf_name, 0, {})
        wait_for_udf_to_exist(as_connection, self.udf_name)

        def teardown():
            """
            Teardown Method, Checks to see if the UDF is on the server,
            if it is: removes it and waits for the removal
            process to complete
            """
            udf_name = TestUdfRemove.udf_name
            udf_list = as_connection.udf_list({"timeout": 180000})

            for udf in udf_list:
                if udf["name"] == udf_name:
                    as_connection.udf_remove(udf_name)
                    wait_for_udf_removal(as_connection, udf_name)
                    break

        request.addfinalizer(teardown)

    def setup_class(cls):
        """
        Setup class, sets the name of the example UDF used in the tests
        """
        cls.udf_name = "example.lua"

    def test_udf_remove_with_no_policy(self):
        """
        Test to verify a proper call to udf_remove will  remove a UDF
        """
        module = "example.lua"

        status = self.as_connection.udf_remove(module)
        assert status == AerospikeStatus.AEROSPIKE_OK

        wait_for_udf_removal(self.as_connection, module)
        udf_list = self.as_connection.udf_list({"timeout": 180000})

        present = False
        for udf in udf_list:
            if "example.lua" == udf["name"]:
                present = True

        assert not present

    @pytest.mark.xfail(reason="This is the only method which allows" + " invalid timeout")
    def test_udf_remove_with_invalid_timeout_policy_value(self):
        """
        Verify that an incorrect timeout policy will not prevent UDF removal
        """
        policy = {"timeout": 0.1}
        module = "example.lua"

        with pytest.raises(e.ParamError):
            status = self.as_connection.udf_remove(module, policy)

        #  Wait for the removal to take place
        wait_for_udf_removal(self.as_connection, module)

        assert status == 0

    def test_udf_remove_with_proper_timeout_policy_value(self):
        """
        Verify that udf_remove with a correct timeout policy argument
        functions.
        """
        policy = {"timeout": 180000}
        module = "example.lua"

        status = self.as_connection.udf_remove(module, policy)

        assert status == AerospikeStatus.AEROSPIKE_OK

        #  Wait for the removal to take place
        wait_for_udf_removal(self.as_connection, module)

        udf_list = self.as_connection.udf_list({"timeout": 180000})

        present = False
        for udf in udf_list:
            if "example.lua" == udf["name"]:
                present = True

        assert not present

    def test_udf_remove_with_unicode_filename(self):
        """
        Test to ensure that unicode filenames may be used to remove UDFs
        """
        policy = {}
        module = "example.lua"
        status = self.as_connection.udf_remove(module, policy)

        assert status == AerospikeStatus.AEROSPIKE_OK

        #  Wait for the removal to take place
        wait_for_udf_removal(self.as_connection, module)

        udf_list = self.as_connection.udf_list({})

        present = False
        for udf in udf_list:
            if "example.lua" == udf["name"]:
                present = True

        assert not present


@pytest.mark.usefixtures("connection_with_udf")
class TestIncorrectCallsToUDFRemove(object):
    """
    These are all tests where udf_remove fails for various reasons,
    So we skip removing and re-adding the UDF before and after each test
    """

    def setup_class(cls):
        """
        setup the class attribute indicating the udf to load
        """
        cls.udf_to_load = "example.lua"

    def test_udf_remove_with_proper_parameters_without_connection(self):
        """
        Test to verify that attempting to remove a UDF before connection
        raises an error
        """
        config = TestBaseClass.get_connection_config()

        client1 = aerospike.client(config)
        client1.close()
        policy = {}
        module = "example.lua"

        with pytest.raises(e.ClusterError) as err_info:
            client1.udf_remove(module, policy)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_CLUSTER_ERROR

    def test_udf_remove_with_non_existent_module(self):
        """
        Test to ensure that the removal of a non existant UDF raises an Error
        """
        policy = {}
        module = "some_fake_module_that_does_not_exist"

        if is_greater_451(self.string_server_version):
            self.as_connection.udf_remove(module, policy)
        else:
            with pytest.raises(e.UDFError):
                self.as_connection.udf_remove(module, policy)

    def test_udf_remove_without_parameters(self):
        """
        Test to verify that udf_remove raises an error with no parameters
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.udf_remove()
        assert "argument 'filename' (pos 1)" in str(typeError.value)

    def test_udf_remove_with_none_as_parameters(self):
        """
        Test to verify that udf_remove raises and error when None is
        used as parameters
        """
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.udf_remove(None, None)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    @pytest.mark.parametrize("udf", [None, False, 1, 1.5, ("cool", "other")])
    def test_udf_with_non_string_module_name(self, udf):
        """
        Tests for incorrect udf module name types
        """
        with pytest.raises(e.ParamError):
            self.as_connection.udf_remove(udf)

    @pytest.mark.xfail(reason="These do not raise errors")
    @pytest.mark.parametrize("policy", [False, "policy", 5, (), []])
    def test_udf_remove_with_invalid_policy_type(self, policy):
        """
        Tests for incorrect policy argument types
        """
        with pytest.raises(TypeError):
            self.as_connection.udf_remove("example.lua", policy)



# Insert 3 records
keys = [("test", "demo", f"employee{i}") for i in range(1, 4)]
bins = [
    {"id": 100, "balance": 200},
    {"id": 101, "balance": 400},
    {"id": 102, "balance": 300}
]
for key, bin in zip(keys, bins):
    client.put(key, bin)

# Apply a user defined function (UDF) to a batch
# of records using batch_apply.
client.udf_put("batch_apply.lua")

args = ["balance", 0.5, 100]
batchRecords = client.batch_apply(keys, "batch_apply", "tax", args)

print(batchRecords.result)
# 0

for batchRecord in batchRecords.batch_records:
    print(f"{batchRecord.result}: {batchRecord.record}")
# 0: (('test', 'demo', 'employee1', bytearray(b'...')), {'ttl': 2592000, 'gen': 2}, {'SUCCESS': 0})
# 0: (('test', 'demo', 'employee2', bytearray(b'...')), {'ttl': 2592000, 'gen': 2}, {'SUCCESS': 100})
# 0: (('test', 'demo', 'employee3', bytearray(b'...')), {'ttl': 2592000, 'gen': 2}, {'SUCCESS': 50})



# -*- coding: utf-8 -*-

import pytest
import time
from . import test_data

from .test_base_class import TestBaseClass
import aerospike
from aerospike import exception as e


class SomeClass(object):
    pass


@pytest.mark.usefixtures("as_connection")
class TestExists:
    @pytest.mark.parametrize("key, record", test_data.pos_data)
    def test_pos_exists_with_diff_datatype(self, key, record, put_data):
        """
        Invoke exists() for diffrent record data.
        """
        put_data(self.as_connection, key, record)
        key, meta = self.as_connection.exists(key)

        assert meta["gen"] is not None
        assert meta["ttl"] is not None

    def test_pos_exists_with_key_and_policy(self, put_data):
        """
        Invoke exists() with a key and policy.
        """
        key = ("test", "demo", 1)
        record = {"Name": "Jeff"}
        policy = {
            "max_retries": 1,
        }

        put_data(self.as_connection, key, record)

        key, meta = self.as_connection.exists(key, policy)

        assert meta["gen"] is not None
        assert meta["ttl"] is not None

    # TODO: duplicate test name
    @pytest.mark.parametrize(
        "key, record, policy",
        [
            (("test", "demo", "p_None"), {"name": "John"}, None),
            (("test", "demo", "p_Replica"), {"name": "Michel"}, {"max_retries": 1}),
            (("test", "demo", "p_consistency_level"), {"name": "Michel"}, {"max_retries": 1}),
            (("test", "demo", "p_consistency_level"), {"name": "Michel"}, {"max_retries": 1}),
        ],
    )
    def test_pos_exists_with_key_and_policy1(self, key, record, policy, put_data):
        """
        Invoke exists() with key and policy.
        """
        put_data(self.as_connection, key, record, _policy=policy)
        key, meta = self.as_connection.exists(key, policy)

        assert meta["gen"] is not None
        assert meta["ttl"] is not None

    # Negative Tests

    def test_neg_exists_with_record_expiry(self, put_data):
        key = ("test", "demo", 30)
        rec = {"name": "John"}
        meta = {"gen": 3, "ttl": 1}
        put_data(self.as_connection, key, rec, meta)
        time.sleep(2)

        key, meta = self.as_connection.exists(key)
        assert meta is None

    @pytest.mark.parametrize(
        "key, ex, ex_code",
        [
            # reason for xfail CLIENT-533
            (("test", "demo", "non-existent"), e.RecordNotFound, 2),  # non-existent key
            # non-existent set
            (("test", "set", 1), e.RecordNotFound, 2),
            (("namespace", "demo", 1), e.ClientError, -1),  # non-existent Namespace
            # None set in key tuple.
            (("test", None, 2), e.RecordNotFound, 2),
            (("test", "demo", "Non_existing_key"), e.RecordNotFound, 2),  # Non_existing_key
        ],
    )
    def test_neg_exists_with_non_existent_data(self, key, ex, ex_code):
        """
        Invoke exists() for non-existent data.
        """
        try:
            key, meta = self.as_connection.exists(key)
            assert meta is None
            """
            We are making the api backward compatible. In case of RecordNotFound an
            exception will not be raised. Instead Ok response is returned withe the
            meta as None. This might change with further releases.
            """
        except ex as exception:
            assert exception.code == ex_code

    def test_neg_exists_with_only_key_without_connection(self):
        """
        Invoke exists() with a key and not policy's dict and no connection
        """
        key = ("test", "demo", 1)
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        try:
            key, _ = client1.exists(key)

        except e.ClusterError as exception:
            assert exception.code == 11

    @pytest.mark.parametrize(
        "key, record, meta, policy",
        [
            (("test", "demo", 20), {"name": "John"}, {"gen": 3, "ttl": 1}, {"total_timeout": 2}),
        ],
    )
    def test_neg_exists_with_low_timeout(self, key, record, meta, policy, put_data):
        try:
            put_data(self.as_connection, key, record, meta, policy)
        except e.TimeoutError as exception:
            assert exception.code == 9
        key, meta = self.as_connection.exists(key)
        assert meta["gen"] is not None

    def test_neg_exists_with_no_paramters(self):
        """
        Invoke self() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.exists()

        assert "argument 'key' (pos 1)" in str(typeError.value)

    @pytest.mark.parametrize(
        "key, record, policy, ex_code, ex_msg",
        [
            # timeout_is_string
            (
                ("test", "demo", "timeout_is_string"),
                {"names": ["John", "Marlen", "Steve"]},
                {"total_timeout": "1000"},
                -2,
                "timeout is invalid",
            ),
            (("test", "demo", "policy_is_string"), {"Name": "Jeff"}, "policy_str", -2, "policy must be a dict"),
        ],
    )
    def test_neg_exists_with_invalid_meta(self, key, record, policy, ex_code, ex_msg, put_data):
        """
        Invoke exists() with a key and timeout as string.
        """
        put_data(self.as_connection, key, record)

        with pytest.raises(e.ParamError):
            key, _ = self.as_connection.exists(key, policy)

        # except e.ParamError as exception:
        #     assert exception.code == ex_code
        #     assert exception.msg == ex_msg

    @pytest.mark.parametrize("key, ex_code, ex_msg", test_data.key_neg)
    def test_neg_exists_key_invalid_data(self, key, ex_code, ex_msg):
        """
        Invoke exists() with invalid key
        """
        with pytest.raises(e.ParamError):
            key, _ = self.as_connection.exists(key)



# -*- coding: utf-8 -*-
import pytest
from .test_base_class import TestBaseClass
from aerospike import exception as e
from .as_status_codes import AerospikeStatus

import aerospike


class TestTouch(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        self.test_demo_1_digest = aerospike.calc_digest("test", "demo", 1)
        self.added_keys = []

        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": i}
            as_connection.put(key, rec)
            self.added_keys.append(key)

        def teardown():
            for key in self.added_keys:
                as_connection.remove(key)
            self.added_keys = []

        request.addfinalizer(teardown)

    def test_touch_with_no_parameters(self):
        """
        Invoke touch() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.touch()
        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_touch_with_correct_paramters(self):
        """
        Invoke touch() with correct parameters
        """
        key = ("test", "demo", 1)
        response = self.as_connection.touch(key, 120)

        assert response == AerospikeStatus.AEROSPIKE_OK

    def test_touch_with_correct_policy(self):
        """
        Invoke touch() with correct policy
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 180000, "retry": aerospike.POLICY_RETRY_ONCE}
        response = self.as_connection.touch(key, 120, {}, policy)
        assert response == AerospikeStatus.AEROSPIKE_OK

    def test_touch_with_policy_key_send(self):
        """
        Invoke touch() with policy key send
        """
        key = ("test", "demo", 1)
        policy = {
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
        }
        self.as_connection.touch(key, 120, {}, policy)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "name1"}
        assert key == ("test", "demo", None, self.test_demo_1_digest)

    def test_touch_with_policy_key_digest(self):
        """
        Invoke touch() with policy key digest
        """
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        rec = {"name": "name%s" % (str(1)), "age": 1, "nolist": [1, 2, 3]}
        self.as_connection.put(key, rec)

        policy = {"key": aerospike.POLICY_KEY_DIGEST, "retry": aerospike.POLICY_RETRY_NONE}
        self.as_connection.touch(key, 120, {}, policy)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "name1", "nolist": [1, 2, 3]}
        assert key == ("test", "demo", None, bytearray(b"asd;as[d'as;djk;uyfl"))
        self.as_connection.remove(key)

    def test_touch_with_policy_key_gen_EQ_ignore(self):
        """
        Invoke touch() with gen eq positive ignore
        """
        key = ("test", "demo", 1)
        policy = {
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_IGNORE,
        }

        meta = {"gen": 10, "ttl": 1200}
        self.as_connection.touch(key, 120, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "name1"}
        assert key == ("test", "demo", None, self.test_demo_1_digest)

    def test_touch_with_policy_key_gen_EQ_positive(self):
        """
        Invoke touch() with gen eq positive
        """
        key = ("test", "demo", 1)
        policy = {
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_EQ,
        }
        (key, meta) = self.as_connection.exists(key)

        gen = meta["gen"]
        meta = {"gen": gen, "ttl": 1200}
        self.as_connection.touch(key, 120, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "name1"}
        assert key == ("test", "demo", None, self.test_demo_1_digest)

    def test_touch_with_policy_key_gen_EQ_not_equal(self):
        """
        Invoke touch() with policy key EQ not equal
        """
        key = ("test", "demo", 1)
        policy = {
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_EQ,
        }
        meta = {"gen": 10, "ttl": 1200}

        with pytest.raises(e.RecordGenerationError) as err_info:
            self.as_connection.touch(key, 120, meta, policy)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_RECORD_GENERATION

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "name1"}
        assert key == ("test", "demo", None, self.test_demo_1_digest)

    def test_touch_with_policy_key_gen_GT_lesser(self):
        """
        Invoke touch() with gen GT lesser
        """
        key = ("test", "demo", 1)
        policy = {
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_GT,
        }
        (key, meta) = self.as_connection.exists(key)

        gen = meta["gen"]
        meta = {"gen": gen, "ttl": 1200}
        with pytest.raises(e.RecordGenerationError) as err_info:
            self.as_connection.touch(key, 120, meta, policy)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_RECORD_GENERATION

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "name1"}
        assert key == ("test", "demo", None, self.test_demo_1_digest)

    def test_touch_with_policy_key_gen_GT_positive(self):
        """
        Invoke touch() with gen GT positive
        """
        key = ("test", "demo", 1)
        policy = {
            "key": aerospike.POLICY_KEY_SEND,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "gen": aerospike.POLICY_GEN_GT,
        }
        (key, meta) = self.as_connection.exists(key)

        gen = meta["gen"]
        meta = {"gen": gen + 5, "ttl": 1200}
        self.as_connection.touch(key, 120, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == {"age": 1, "name": "name1"}
        assert key == ("test", "demo", None, self.test_demo_1_digest)

    def test_touch_with_incorrect_policy(self):
        """
        Invoke touch() with incorrect policy
        """
        key = ("test", "demo", 1)
        policy = {"total_timeout": 0.5}

        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.touch(key, 120, {}, policy)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_touch_with_nonexistent_key(self):
        """
        Invoke touch() with non-existent key
        """
        key = ("test", "demo", 1000)

        with pytest.raises(e.RecordNotFound) as err_info:
            self.as_connection.touch(key, 120)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_RECORD_NOT_FOUND

    def test_touch_value_string(self):
        """
        Invoke touch() not a string
        """
        key = ("test", "demo", 1)
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.touch(key, "name")

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_touch_with_extra_parameter(self):
        """
        Invoke touch() with extra parameter.
        """
        key = ("test", "demo", 1)
        policy = {}
        with pytest.raises(TypeError) as typeError:
            self.as_connection.touch(key, 120, {}, policy, "")

        assert "touch() takes at most 4 arguments (5 given)" in str(typeError.value)

    def test_touch_policy_is_string(self):
        """
        Invoke touch() with policy is string
        """
        key = ("test", "demo", 1)
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.touch(key, 120, {}, "")

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_touch_with_correct_paramters_without_connection(self):
        """
        Invoke touch() with correct parameters without connection
        """
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()
        key = ("test", "demo", 1)

        with pytest.raises(e.ClusterError) as err_info:
            client1.touch(key, 120)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_CLUSTER_ERROR

    def test_touch_withttlvalue_greaterthan_maxsize(self):
        """
        Invoke touch() with ttl value greater than (2^63-1)
        """
        key = ("test", "demo", 1)
        meta = {"gen": 10, "ttl": 2**64}
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.touch(key, 120, meta, None)

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM



# Error Codes and messages for various exceptions

"""
Creates Classes so that we can access them like: cls.err_type.err_code
Roughly maps to: exception_types.h
"""


class AerospikeStatus(object):
    AEROSPIKE_OK = 0
    AEROSPIKE_SERVER_ERROR = -10
    AEROSPIKE_INVALID_HOST = -4
    AEROSPIKE_ERR_CLIENT = -1
    AEROSPIKE_ERR_PARAM = -2
    AEROSPIKE_ERR_RECORD_NOT_FOUND = 2
    AEROSPIKE_ERR_RECORD_GENERATION = 3
    AEROSPIKE_ERR_REQUEST_INVALID = 4
    AEROSPIKE_CLUSTER_ERROR = 11
    AEROSPIKE_ERR_BIN_INCOMPATIBLE_TYPE = 12
    AEROSPIKE_ERR_NAMESPACE_NOT_FOUND = 20
    AEROSPIKE_ERR_UDF = 100
    AEROSPIKE_ERR_INDEX_NOT_FOUND = 201
    LUA_FILE_NOT_FOUND = 1302
    AEROSPIKE_ERR_MAX_RETRIES_EXCEEDED = -12



# -*- coding: utf-8 -*-
import pytest
from .test_base_class import TestBaseClass
from aerospike_helpers.operations import list_operations, operations

import aerospike
from aerospike import exception as e

# OPERATIONS
# aerospike.OPERATOR_WRITE
# aerospike.OPERATOR_APPEND
# aerospike.OPERATOR_PREPEND
# aerospike.OPERATOR_INCR
# aerospike.OPERATOR_READ
# aerospike.OPERATOR_TOUCH
# aerospike.OP_LIST_APPEND
# aerospike.OP_LIST_APPEND_ITEMS
# aerospike.OP_LIST_INSERT
# aerospike.OP_LIST_INSERT_ITEMS
# aerospike.OP_LIST_POP
# aerospike.OP_LIST_POP_RANGE
# aerospike.OP_LIST_REMOVE
# aerospike.OP_LIST_REMOVE_RANGE
# aerospike.OP_LIST_CLEAR
# aerospike.OP_LIST_SET
# aerospike.OP_LIST_GET
# aerospike.OP_LIST_GET_RANGE
# aerospike.OP_LIST_TRIM
# aerospike.OP_LIST_SIZE
# aerospike.OP_MAP_SET_POLICY
# aerospike.OP_MAP_PUT
# aerospike.OP_MAP_PUT_ITEMS
# aerospike.OP_MAP_INCREMENT
# aerospike.OP_MAP_DECREMENT
# aerospike.OP_MAP_SIZE
# aerospike.OP_MAP_CLEAR
# aerospike.OP_MAP_REMOVE_BY_KEY
# aerospike.OP_MAP_REMOVE_BY_KEY_LIST
# aerospike.OP_MAP_REMOVE_BY_KEY_RANGE
# aerospike.OP_MAP_REMOVE_BY_VALUE
# aerospike.OP_MAP_REMOVE_BY_VALUE_LIST
# aerospike.OP_MAP_REMOVE_BY_VALUE_RANGE
# aerospike.OP_MAP_REMOVE_BY_INDEX
# aerospike.OP_MAP_REMOVE_BY_INDEX_RANGE
# aerospike.OP_MAP_REMOVE_BY_RANK
# aerospike.OP_MAP_REMOVE_BY_RANK_RANGE
# aerospike.OP_MAP_GET_BY_KEY
# aerospike.OP_MAP_GET_BY_KEY_RANGE
# aerospike.OP_MAP_GET_BY_VALUE
# aerospike.OP_MAP_GET_BY_VALUE_RANGE
# aerospike.OP_MAP_GET_BY_INDEX
# aerospike.OP_MAP_GET_BY_INDEX_RANGE
# aerospike.OP_MAP_GET_BY_RANK
# aerospike.OP_MAP_GET_BY_RANK_RANGE


class TestOperate(object):
    def setup_class(cls):
        """
        Setup class.
        """
        cls.client_no_typechecks = TestBaseClass.get_new_connection({"strict_types": False})

    def teardown_class(cls):
        TestOperate.client_no_typechecks.close()

    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Setup Method
        """
        keys = []

        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": i}
            as_connection.put(key, rec)

        key = ("test", "demo", 6)
        rec = {"age": 6.3}
        as_connection.put(key, rec)
        keys.append(key)

        key = ("test", "demo", "bytearray_key")
        rec = {"bytearray_bin": bytearray("asd;as[d'as;d", "utf-8")}
        as_connection.put(key, rec)

        keys.append(key)

        key = ("test", "demo", "bytes_key")
        rec = {"bytes_bin": b""}
        as_connection.put(key, rec)

        keys.append(key)

        key = ("test", "demo", "list_key")
        rec = {"int_bin": [1, 2, 3, 4], "string_bin": ["a", "b", "c", "d"]}
        as_connection.put(key, rec)

        keys.append(key)

        key = ("test", "demo", "existing_key")
        rec = {"dict": {"a": 1}, "bytearray": bytearray("abc", "utf-8"), "float": 3.4, "list": ["a"]}
        as_connection.put(key, rec)

        keys.append(key)

        def teardown():
            """
            Teardown Method
            """
            for i in range(5):
                key = ("test", "demo", i)
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass
            for key in keys:
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "key, llist, expected",
        [
            (
                ("test", "demo", 1),
                [operations.prepend("name", "ram"), operations.increment("age", 3), operations.read("name")],
                {"name": "ramname1"},
            ),
            (
                ("test", "demo", 1),  # with_write_float_value
                [operations.write("write_bin", {"no": 89.8}), operations.read("write_bin")],
                {"write_bin": {"no": 89.8}},
            ),
            (
                ("test", "demo", 1),  # write positive
                [operations.write("write_bin", {"no": 89}), operations.read("write_bin")],
                {"write_bin": {"no": 89}},
            ),
            (
                ("test", "demo", 1),  # write_tuple_positive
                [operations.write("write_bin", ("a", "b", "c")), operations.read("write_bin")],
                {"write_bin": ("a", "b", "c")},
            ),
            (
                ("test", "demo", 1),  # with_bin_bytearray
                [operations.prepend("asd[;asjk", "ram"), operations.read("asd[;asjk")],
                {"asd[;asjk": "ram"},
            ),
            (
                ("test", "demo", "bytearray_key"),  # with_operator append_val bytearray
                [operations.append("bytearray_bin", bytearray("abc", "utf-8")), operations.read("bytearray_bin")],
                {"bytearray_bin": bytearray("asd;as[d'as;dabc", "utf-8")},
            ),
            (
                ("test", "demo", "bytearray_new"),  # with_operator append_val bytearray_newrecord
                [
                    operations.append("bytearray_bin", bytearray("asd;as[d'as;d", "utf-8")),
                    operations.read("bytearray_bin"),
                ],
                {"bytearray_bin": bytearray("asd;as[d'as;d", "utf-8")},
            ),
            (
                ("test", "demo", "bytes_key"),  # with_operator append_val bytes
                [operations.append("bytes_bin", b"abc"), operations.read("bytes_bin")],
                {"bytes_bin": b"abc"},
            ),
            (
                ("test", "demo", "bytes_new"),  # with_operator append_val bytes_newrecord
                [operations.append("bytes_bin", b"asd;as[d'as;d"), operations.read("bytes_bin")],
                {"bytes_bin": b"asd;as[d'as;d"},
            ),
            (
                ("test", "demo", "bytearray_key"),  # with_operatorprepend_valbytearray
                [operations.prepend("bytearray_bin", bytearray("abc", "utf-8")), operations.read("bytearray_bin")],
                {"bytearray_bin": bytearray("abcasd;as[d'as;d", "utf-8")},
            ),
            (
                ("test", "demo", "bytearray_new"),  # with_operatorprepend_valbytearray_newrecord
                [
                    operations.prepend("bytearray_bin", bytearray("asd;as[d'as;d", "utf-8")),
                    operations.read("bytearray_bin"),
                ],
                {"bytearray_bin": bytearray("asd;as[d'as;d", "utf-8")},
            ),
            (
                ("test", "demo", "bytes_key"),  # with_operator prepend_val bytes
                [operations.prepend("bytes_bin", b"abc"), operations.read("bytes_bin")],
                {"bytes_bin": b"abc"},
            ),
            (
                ("test", "demo", "bytes_new"),  # with_operator prepend_val bytes_newrecord
                [operations.prepend("bytes_bin", b"asd;as[d'as;d"), operations.read("bytes_bin")],
                {"bytes_bin": b"asd;as[d'as;d"},
            ),
            (
                ("test", "demo", 1),  # write_bool_positive
                [operations.write("write_bin", True), operations.read("write_bin")],
                {"write_bin": 1},
            ),
            (
                ("test", "demo", 1),  # write_bool_positive
                [operations.write("write_bin", False), operations.read("write_bin")],
                {"write_bin": 0},
            ),
        ],
    )
    def test_pos_operate_with_correct_paramters(self, key, llist, expected):
        """
        Invoke operate() with correct parameters
        """

        key, _, bins = self.as_connection.operate(key, llist)

        assert bins == expected
        self.as_connection.remove(key)

    @pytest.mark.parametrize(
        "key, llist, expected",
        [
            (
                ("test", "demo", 1),
                [
                    operations.write("write_bin", {"no": 89.8}),
                    operations.write("write_bin2", {"no": 100}),
                    operations.delete(),
                ],
                {},
            ),
        ],
    )
    def test_pos_operate_delete_with_correct_paramters(self, key, llist, expected):
        """
        Invoke operate() with correct parameters
        """

        key, _, bins = self.as_connection.operate(key, llist)

        assert bins == expected

    def test_pos_operate_with_increment_positive_float_value(self):
        """
        Invoke operate() with correct parameters
        """
        if TestOperate.skip_old_server is True:
            pytest.skip("Server does not support increment on float type")
        key = ("test", "demo", 6)
        llist = [operations.increment("age", 3.5), operations.read("age")]

        _, _, bins = self.as_connection.operate(key, llist)

        assert bins == {"age": 9.8}

    def test_pos_operate_with_correct_policy(self):
        """
        Invoke operate() with correct policy
        """
        key = ("test", "demo", 1)
        policy = {
            "timeout": 180000,
            "key": aerospike.POLICY_KEY_SEND,
            "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
        }

        llist = [operations.append("name", "aa"), operations.increment("age", 3), operations.read("name")]

        _, _, bins = self.as_connection.operate(key, llist, {}, policy)

        assert bins == {"name": "name1aa"}

        self.as_connection.remove(key)

    @pytest.mark.parametrize(
        "key, policy, meta, llist",
        [
            (
                ("test", "demo", 1),
                {
                    "key": aerospike.POLICY_KEY_SEND,
                    "gen": aerospike.POLICY_GEN_IGNORE,
                    "commit_level": aerospike.POLICY_COMMIT_LEVEL_ALL,
                },
                {"gen": 10, "ttl": 1200},
                [operations.append("name", "aa"), operations.increment("age", 3), operations.read("name")],
            ),
        ],
    )
    def test_pos_operate_with_policy_gen_ignore(self, key, policy, meta, llist):
        """
        Invoke operate() with gen ignore.
        """

        key, meta, bins = self.as_connection.operate(key, llist, meta, policy)

        assert bins == {"name": "name1aa"}

    def test_pos_operate_with_policy_gen_EQ(self):
        """
        Invoke operate() with gen EQ positive.
        """
        key = ("test", "demo", 1)
        policy = {"gen": aerospike.POLICY_GEN_EQ}
        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen}

        llist = [operations.append("name", "aa"), operations.increment("age", 3), operations.read("name")]
        (key, meta, bins) = self.as_connection.operate(key, llist, meta, policy)

        assert bins == {"name": "name1aa"}

    @pytest.mark.parametrize(
        "key, llist", [(("test", "demo", 1), [operations.touch(4000)]), (("test", "demo", 1), [operations.touch(4000)])]
    )
    def test_pos_operate_touch_operation_with_bin_and_value_combination(self, key, llist):
        """
        Invoke operate() with touch value with bin and value combination.
        """

        self.as_connection.operate(key, llist)

        (key, meta) = self.as_connection.exists(key)

        assert meta["ttl"] is not None

    def test_pos_operate_with_policy_gen_EQ_not_equal(self):
        """
        Invoke operate() with a mismatched generation, and verify
        it does not succeed
        """
        key = ("test", "demo", 1)
        policy = {"gen": aerospike.POLICY_GEN_EQ}

        (_, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {
            "gen": gen + 5,
        }
        llist = [
            operations.append("name", "aa"),
            operations.increment("age", 3),
        ]

        with pytest.raises(e.RecordGenerationError):
            self.as_connection.operate(key, llist, meta, policy)

        _, _, bins = self.as_connection.get(key)
        assert bins == {"age": 1, "name": "name1"}

    def test_pos_operate_with_policy_gen_GT_mismatch(self):
        """
        Invoke operate() with gen GT policy, with amatching generation
        then verify that the operation was rejected properly
        """
        key = ("test", "demo", 1)
        policy = {"gen": aerospike.POLICY_GEN_GT}
        _, meta = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen}

        llist = [operations.append("name", "aa"), operations.increment("age", 3)]

        with pytest.raises(e.RecordGenerationError):
            self.as_connection.operate(key, llist, meta, policy)

        _, _, bins = self.as_connection.get(key)
        assert bins == {"age": 1, "name": "name1"}

    def test_pos_operate_touch_with_meta(self):
        """
        Invoke operate() OPERATE_TOUCH using meta to pass in ttl.
        """
        key = ("test", "demo", 1)
        meta = {"ttl": 1200}

        llist = [operations.touch()]

        self.as_connection.operate(key, llist, meta)

        _, meta = self.as_connection.exists(key)

        assert meta["ttl"] <= 1200 and meta["ttl"] >= 1150

    def test_pos_operate_with_policy_gen_GT(self):
        """
        Invoke operate() with gen GT positive.
        """
        key = ("test", "demo", 1)
        policy = {"gen": aerospike.POLICY_GEN_GT}
        _, meta = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen + 5}

        llist = [operations.append("name", "aa"), operations.increment("age", 3), operations.read("name")]
        _, _, bins = self.as_connection.operate(key, llist, meta, policy)

        assert bins == {"name": "name1aa"}

    def test_pos_operate_with_nonexistent_key(self):
        """
        Invoke operate() with non-existent key
        """
        new_key = ("test", "demo", "key11")
        llist = [operations.prepend("loc", "mumbai"), operations.read("loc")]
        _, _, bins = self.as_connection.operate(new_key, llist)

        assert bins == {"loc": "mumbai"}
        self.as_connection.remove(new_key)

    def test_pos_operate_with_nonexistent_bin(self):
        """
        Invoke operate() with non-existent bin
        """
        key = ("test", "demo", 1)
        llist = [operations.append("addr", "pune"), operations.read("addr")]
        _, _, bins = self.as_connection.operate(key, llist)

        assert bins == {"addr": "pune"}

    def test_pos_operate_increment_nonexistent_key(self):
        """
        Invoke operate() with increment with nonexistent_key
        """
        key = ("test", "demo", "non_existentkey")
        llist = [operations.increment("age", 5)]

        self.as_connection.operate(key, llist)

        _, _, bins = self.as_connection.get(key)

        assert bins == {"age": 5}

        self.as_connection.remove(key)

    def test_pos_operate_with_correct_paramters_without_connection(self):
        """
        Invoke operate() with correct parameters without connection
        """
        key = ("test", "demo", 1)
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()
        llist = [operations.touch()]

        with pytest.raises(e.ClusterError):
            client1.operate(key, llist)

    def test_pos_operate_write_set_to_aerospike_null(self):
        """
        Invoke operate() with write command with bin set to aerospike_null
        """
        key = ("test", "demo", "null_record")

        bins = {"name": "John", "no": 3}

        self.as_connection.put(key, bins)

        llist = [operations.write("no", aerospike.null()), operations.read("no")]

        _, _, bins = self.as_connection.operate(key, llist)
        assert {} == bins

        self.as_connection.remove(key)

    @pytest.mark.parametrize(
        "key, llist, expected",
        [
            (
                ("test", "demo", "prepend_int"),  # prepend_with_int
                [operations.prepend("age", 4), operations.read("age")],
                {"age": 4},
            ),
            (
                ("test", "demo", "append_dict"),  # append_with_dict
                [operations.append("dict", {"a": 1, "b": 2}), operations.read("dict")],
                {"dict": {"a": 1, "b": 2}},
            ),
            (
                ("test", "demo", "incr_string"),  # incr_with_string
                [operations.increment("name", "aerospike"), operations.read("name")],
                {"name": "aerospike"},
            ),
        ],
    )
    def test_pos_operate_new_record(self, key, llist, expected):
        """
        Invoke operate() with prepend command on a new record
        """
        _, _, bins = TestOperate.client_no_typechecks.operate(key, llist)
        assert expected == bins
        TestOperate.client_no_typechecks.remove(key)

    @pytest.mark.parametrize(
        "key, llist",
        [
            (("test", "demo", 1), [operations.prepend("age", 4), operations.read("age")]),
            (
                ("test", "demo", "existing_key"),  # Existing list
                [operations.prepend("list", ["c"]), operations.read("list")],
            ),
            (
                ("test", "demo", "existing_key"),  # Existing dict
                [operations.append("dict", {"c": 2}), operations.read("dict")],
            ),
            (
                ("test", "demo", "existing_key"),  # Exiting float
                [operations.append("float", 3.4), operations.read("float")],
            ),
            (
                ("test", "demo", 1),  # Existing string
                [operations.increment("name", "aerospike"), operations.read("name")],
            ),
            (
                ("test", "demo", "existing_key"),  # Existing Bytearray
                [operations.increment("bytearray", bytearray("abc", "utf-8")), operations.read("bytearray")],
            ),
        ],
    )
    def test_pos_operate_prepend_with_existing_record(self, key, llist):
        """
        Invoke operate() with prepend command on a existing record
        """

        with pytest.raises(e.BinIncompatibleType):
            TestOperate.client_no_typechecks.operate(key, llist)

        TestOperate.client_no_typechecks.remove(key)

    def test_pos_operate_incr_with_geospatial_new_record(self):
        """
        Invoke operate() with incr command on a new record
        """
        key = ("test", "demo", "geospatial_key")

        llist = [
            operations.increment("geospatial", aerospike.GeoJSON({"type": "Point", "coordinates": [42.34, 58.62]})),
            operations.read("geospatial"),
        ]

        _, _, bins = TestOperate.client_no_typechecks.operate(key, llist)

        assert bins["geospatial"].unwrap() == {"coordinates": [42.34, 58.62], "type": "Point"}
        TestOperate.client_no_typechecks.remove(key)

    def test_pos_operate_with_bin_length_extra_nostricttypes(self):
        """
        Invoke operate() with bin length extra. Strict types disabled
        """
        key = ("test", "demo", 1)

        max_length = "a" * 21

        llist = [operations.prepend("name", "ram"), operations.increment(max_length, 3)]

        TestOperate.client_no_typechecks.operate(key, llist)

        _, _, bins = TestOperate.client_no_typechecks.get(key)

        assert bins == {"name": "ramname1", "age": 1}

    def test_pos_operate_prepend_set_to_aerospike_null(self):
        """
        Invoke operate() with prepend command with bin set to aerospike_null
        """
        key = ("test", "demo", "null_record")

        bins = {"name": "John", "no": 3}

        assert 0 == self.as_connection.put(key, bins)

        (key, _, bins) = self.as_connection.get(key)

        assert {"name": "John", "no": 3} == bins

        llist = [operations.prepend("no", aerospike.null()), operations.read("no")]

        try:
            (key, _, bins) = self.as_connection.operate(key, llist)

        except e.InvalidRequest as exception:
            assert exception.code == 4
        self.as_connection.remove(key)

    @pytest.mark.parametrize(
        "list, result, bin, expected",
        [
            (
                [list_operations.list_append("int_bin", 7), list_operations.list_get("int_bin", 4)],
                {"int_bin": 7},
                "int_bin",
                [1, 2, 3, 4, 7],
            ),
            (
                [
                    list_operations.list_append_items("int_bin", [7, 9]),
                    list_operations.list_get_range("int_bin", 3, 3),
                ],
                {"int_bin": [4, 7, 9]},
                "int_bin",
                [1, 2, 3, 4, 7, 9],
            ),
            (
                [list_operations.list_insert("int_bin", 2, 7), list_operations.list_pop("int_bin", 2)],
                {"int_bin": 7},
                "int_bin",
                [1, 2, 3, 4],
            ),
            (
                [
                    list_operations.list_insert_items("int_bin", 2, [7, 9]),
                    list_operations.list_pop_range("int_bin", 2, 2),
                ],
                {"int_bin": [7, 9]},
                "int_bin",
                [1, 2, 3, 4],
            ),
            (
                [list_operations.list_set("int_bin", 2, 18), list_operations.list_get("int_bin", 2)],
                {"int_bin": 18},
                "int_bin",
                [1, 2, 18, 4],
            ),
            (
                [list_operations.list_set("int_bin", 6, 10), list_operations.list_get("int_bin", 6)],
                {"int_bin": 10},
                "int_bin",
                [1, 2, 3, 4, None, None, 10],  # Inserting outside of the range adds nils in between
            ),
        ],
    )
    def test_pos_operate_with_list_addition_operations(self, list, result, bin, expected):
        """
        Invoke operate() with list addition operations
        """
        key = ("test", "demo", "list_key")

        key, _, bins = self.as_connection.operate(key, list)

        assert bins == result

        key, _, bins = self.as_connection.get(key)
        assert bins[bin] == expected

    @pytest.mark.parametrize(
        "list, bin, expected",
        [
            ([list_operations.list_remove("int_bin", 2)], "int_bin", [1, 2, 4]),
            ([list_operations.list_remove_range("int_bin", 2, 2)], "int_bin", [1, 2]),
            ([list_operations.list_trim("int_bin", 2, 2)], "int_bin", [3, 4]),
            ([list_operations.list_clear("int_bin")], "int_bin", []),
        ],
    )
    def test_pos_operate_with_list_remove_operations(self, list, bin, expected):
        """
        Invoke operate() with list remove operations
        """
        key = ("test", "demo", "list_key")

        self.as_connection.operate(key, list)

        key, _, bins = self.as_connection.get(key)

        assert bins[bin] == expected

    def test_pos_operate_with_list_size(self):
        """
        Invoke operate() with list_size operation
        """
        key = ("test", "demo", "list_key")
        list = [list_operations.list_size("int_bin")]

        key, _, bins = self.as_connection.operate(key, list)

        assert bins == {"int_bin": 4}

    def test_list_increment_with_valid_value(self):
        """
        previous list was [1, 2, 3, 4]
        new should be [1, 2, 23, 4]
        """
        key = ("test", "demo", "list_key")
        list = [list_operations.list_increment("int_bin", 2, 20)]

        _, _, bins = self.as_connection.operate(key, list)

        assert bins == {"int_bin": 23}
        _, _, bins = self.as_connection.get(key)

        assert bins["int_bin"] == [1, 2, 23, 4]

    def test_list_increment_with_incorrect_value_type(self):
        """
        previous list was [1, 2, 3, 4]
        new should be [1, 2, 23, 4]
        """
        key = ("test", "demo", "list_key")
        list = [list_operations.list_increment("int_bin", 2, "twenty")]

        with pytest.raises(e.AerospikeError):
            self.as_connection.operate(key, list)

    def test_pos_operate_with_list_get_range_val_out_of_bounds(self):
        """
        Invoke operate() with list_get_range operation and value out of bounds
        """
        key = ("test", "demo", "list_key")
        list = [list_operations.list_get_range("int_bin", 2, 9)]

        _, _, bins = self.as_connection.operate(key, list)

        assert bins == {"int_bin": [3, 4]}

    def test_pos_operate_with_list_trim_val_with_negative_value(self):
        """
        Invoke operate() with list_trimoperation and value is negative
        """
        key = ("test", "demo", "list_key")
        list = [list_operations.list_trim("int_bin", 1, -9)]

        self.as_connection.operate(key, list)

        _, _, bins = self.as_connection.get(key)

        assert bins["int_bin"] == [2, 3, 4]

    def test_pos_operate_with_list_insert_index_negative(self):
        """
        Invoke operate() with list_insert and item index is a negative value
        """
        key = ("test", "demo", "list_key")
        list = [list_operations.list_insert("int_bin", -2, 9)]

        self.as_connection.operate(key, list)

        _, _, bins = self.as_connection.get(key)

        assert bins["int_bin"] == [1, 2, 9, 3, 4]

    @pytest.mark.parametrize(
        "list, result, bin, expected",
        [
            (
                [list_operations.list_append("string_bin", {"new_val": 1}), list_operations.list_get("string_bin", 4)],
                {"string_bin": {"new_val": 1}},
                "string_bin",
                ["a", "b", "c", "d", {"new_val": 1}],
            ),
            (
                [
                    list_operations.list_append_items("string_bin", [["z", "x"], ("y", "w")]),
                    list_operations.list_get_range("string_bin", 3, 3),
                ],
                {"string_bin": ["d", ["z", "x"], ("y", "w")]},
                "string_bin",
                ["a", "b", "c", "d", ["z", "x"], ("y", "w")],
            ),
            (
                [list_operations.list_insert("string_bin", 2, True), list_operations.list_pop("string_bin", 2)],
                {"string_bin": True},
                "string_bin",
                ["a", "b", "c", "d"],
            ),
            (
                [
                    list_operations.list_insert_items("string_bin", 2, [bytearray("abc", "utf-8"), "xyz"]),
                    list_operations.list_pop_range("string_bin", 2, 2),
                ],
                {"string_bin": [bytearray(b"abc"), "xyz"]},
                "string_bin",
                ["a", "b", "c", "d"],
            ),
        ],
    )
    def test_pos_operate_with_list_operations_different_datatypes(self, list, result, bin, expected):
        """
        Invoke operate() with list operations using different datatypes
        """
        key = ("test", "demo", "list_key")

        _, _, bins = self.as_connection.operate(key, list)

        assert bins == result

        _, _, bins = self.as_connection.get(key)

        assert bins[bin] == expected

    # Negative Tests
    def test_neg_operate_with_no_parameters(self):
        """
        Invoke operate() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.operate()
        assert "key" in str(typeError.value)

    @pytest.mark.parametrize(
        "key, llist, expected",
        [
            (
                ("test", "demo", "bad_key"),
                [
                    operations.delete(),
                ],
                e.RecordNotFound,
            )
        ],
    )
    def test_pos_operate_delete_with_incorrect_paramters(self, key, llist, expected):
        """
        Invoke operate() with correct parameters
        """

        with pytest.raises(expected):
            self.as_connection.operate(key, llist)

    def test_neg_operate_list_operation_bin_notlist(self):
        """
        Invoke operate() with a list operation and bin does not contain list
        """
        key = ("test", "demo", 1)
        list = [list_operations.list_insert("age", 2, 9)]

        with pytest.raises(e.BinIncompatibleType):
            self.as_connection.operate(key, list)

    def test_neg_operate_append_items_not_a_list(self):
        """
        Invoke operate() with list addition operations negative
        """
        key = ("test", "demo", 1)
        ops = [list_operations.list_append_items("int_bin", 7)]
        with pytest.raises(e.ParamError):
            self.as_connection.operate(key, ops)

    @pytest.mark.parametrize(
        "list",
        [
            ([list_operations.list_get("int_bin", 7)]),
            ([list_operations.list_clear("int_bin"), list_operations.list_pop("int_bin", 2)]),
            ([list_operations.list_clear("int_bin"), list_operations.list_remove("int_bin", 2)]),
        ],
    )
    def test_neg_operate_list_invalid_requests(self, list):
        """
        Invoke operate() with list addition operations negative
        """
        key = ("test", "demo", "list_key")
        with pytest.raises(e.OpNotApplicable):
            self.as_connection.operate(key, list)

    def test_neg_operate_with_bin_length_extra(self):
        """
        Invoke operate() with bin length extra. Strict types enabled
        """
        key = ("test", "demo", 1)

        max_length = "a" * 21

        llist = [operations.prepend("name", "ram"), operations.increment(max_length, 3)]

        with pytest.raises(e.BinNameError):
            self.as_connection.operate(key, llist)

    def test_neg_operate_empty_string_key(self):
        """
        Invoke operate() with empty string key
        """
        llist = [operations.prepend("name", "ram")]
        with pytest.raises(e.ParamError):
            self.as_connection.operate("", llist)

    def test_neg_operate_with_extra_parameter(self):
        """
        Invoke operate() with extra parameter.
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 180000}
        llist = [operations.prepend("name", "ram")]
        with pytest.raises(TypeError):
            self.as_connection.operate(key, llist, {}, policy, "")

    def test_neg_operate_policy_is_string(self):
        """
        Invoke operate() with policy is string
        """
        key = ("test", "demo", 1)
        llist = [operations.prepend("name", "ram")]
        try:
            self.as_connection.operate(key, llist, {}, "")

        except e.ParamError as exception:
            assert exception.code == -2

    def test_neg_operate_key_is_none(self):
        """
        Invoke operate() with key is none
        """
        llist = [operations.prepend("name", "ram")]
        try:
            self.as_connection.operate(None, llist)

        except e.ParamError as exception:
            assert exception.code == -2

    def test_neg_operate_append_value_integer(self):
        """
        Invoke operate() with append value is of type integer
        """
        key = ("test", "demo", 1)
        llist = [operations.append("name", 12)]

        try:
            self.as_connection.operate(key, llist)
        except e.ParamError as exception:
            assert exception.code == -2

    def test_neg_operate_with_incorrect_polic(self):
        """
        Invoke operate() with incorrect policy
        """
        key = ("test", "demo", 1)
        policy = {"total_timeout": 0.5}
        llist = [operations.prepend("name", "ram")]

        with pytest.raises(e.ParamError):
            self.as_connection.operate(key, llist, {}, policy)



# -*- coding: utf-8 -*-

import pytest
import time
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


class TestGrantRoles(TestBaseClass):

    pytestmark = pytest.mark.skipif(
        not TestBaseClass.auth_in_use(), reason="No user specified, may be not secured cluster."
    )

    def setup_method(self, method):
        """
        Setup method
        """
        config = TestBaseClass.get_connection_config()
        TestGrantRoles.Me = self
        self.client = aerospike.client(config).connect(config["user"], config["password"])

        try:
            self.client.admin_drop_user("example-test")
            time.sleep(1)
        except e.InvalidUser:
            pass
        policy = {}
        user = "example-test"
        password = "foo2"
        roles = ["read-write"]

        try:
            self.client.admin_create_user(user, password, roles, policy)
            time.sleep(1)
        except e.UserExistsError:
            pass
        self.delete_users = []

    def teardown_method(self, method):
        """
        Teardown method
        """

        policy = {}
        try:
            self.client.admin_drop_user("example-test", policy)
            time.sleep(1)
        except e.InvalidUser:
            pass
        self.client.close()

    def test_grant_roles_without_any_parameters(self):

        with pytest.raises(TypeError):
            self.client.admin_grant_roles()

    def test_grant_roles_with_proper_parameters(self):

        policy = {"timeout": 1000}
        user = "example-test"
        roles = ["read", "read-write", "sys-admin"]

        status = self.client.admin_grant_roles(user, roles, policy)
        assert status == 0
        time.sleep(2)

        user_details = self.client.admin_query_user(user, policy)

        assert user_details == ["read", "read-write", "sys-admin"]

    def test_grant_roles_with_proper_parameters_without_policy(self):

        policy = {"timeout": 1000}
        user = "example-test"
        roles = ["read", "read-write", "sys-admin"]

        status = self.client.admin_grant_roles(user, roles)
        assert status == 0
        time.sleep(2)

        user_details = self.client.admin_query_user(user, policy)

        assert user_details == ["read", "read-write", "sys-admin"]

    def test_grant_roles_with_invalid_timeout_policy_value(self):

        policy = {"timeout": 0.1}
        user = "example-test"
        roles = ["sys-admin"]

        try:
            self.client.admin_grant_roles(user, roles, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"

    def test_grant_roles_with_proper_timeout_policy_value(self):

        user = "example-test"
        roles = ["read-write", "sys-admin"]

        status = self.client.admin_grant_roles(user, roles)

        time.sleep(2)

        assert status == 0

        user_details = self.client.admin_query_user(user)

        assert user_details == ["read-write", "sys-admin"]

    def test_grant_roles_with_none_username(self):

        policy = {"timeout": 20}
        user = None
        roles = ["sys-admin"]

        try:
            self.client.admin_grant_roles(user, roles, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Username should be a string"

    def test_grant_roles_with_empty_username(self):

        policy = {"timeout": 1000}
        user = ""
        roles = ["read-write"]

        try:
            self.client.admin_grant_roles(user, roles, policy)

        except e.InvalidUser as exception:
            assert exception.code == 60
            assert exception.msg == "AEROSPIKE_INVALID_USER"

    def test_grant_roles_with_special_characters_in_username(self):

        policy = {"timeout": 1000}
        user = "!#Q#AEQ@#$%&^*((^&*~~~````["
        password = "abcd"
        roles = ["read-write"]

        try:
            self.client.admin_create_user(user, password, roles, policy)
            time.sleep(1)
        except e.UserExistsError:
            pass

        roles = ["read"]
        status = self.client.admin_grant_roles(user, roles, policy)

        time.sleep(2)

        assert status == 0

        user_details = self.client.admin_query_user(user)

        assert set(user_details) == set(["read", "read-write"])

        self.client.admin_drop_user(user)

    def test_grant_roles_with_empty_roles_list(self):

        policy = {"timeout": 1000}
        user = "example-test"
        roles = []

        try:
            self.client.admin_grant_roles(user, roles, policy)

        except e.InvalidRole as exception:
            assert exception.code == 70
            assert exception.msg == "AEROSPIKE_INVALID_ROLE"

    # TODO: incorrect test
    def test_grant_roles_with_role_name_exceeding_max_length(self):

        policy = {"timeout": 1000}  # noqa: F841
        user = "example-test"
        roles = ["read" * 25, "read-write" * 25]

        with pytest.raises(e.ClientError):
            self.client.admin_grant_roles(user, roles)



# -*- coding: utf-8 -*-

import pytest
import json
import marshal
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


class SomeClass(object):

    pass


def serialize_function_old_server(val):
    return marshal.dumps(val)


def serialize_function(val):
    print("seralize this: {0}".format(val))
    return json.dumps(val)


def client_serialize_function(val):
    print("seralize this: {0}".format(val))
    try:
        return json.dumps(val)
    except Exception:
        print("INVALID JSON")


def deserialize_function(val):
    return json.loads(val)


def deserialize_function_old_server(val):
    return marshal.loads(val)


def client_deserialize_function(val):
    return json.loads(val)


class TestUserSerializer(object):
    def setup_class(cls):
        """
        Setup class
        """
        cls.client = TestBaseClass.get_new_connection()

        TestUserSerializer.skip_old_server = True
        versioninfo = TestUserSerializer.client.info("version")
        for keys in versioninfo:
            for value in versioninfo[keys]:
                if value is not None:
                    versionlist = value[value.find("build") + 6 : value.find("\n")].split(".")
                    if int(versionlist[0]) >= 3 and int(versionlist[1]) >= 6:
                        TestUserSerializer.skip_old_server = False

    def teardown_class(cls):
        TestUserSerializer.client.close()
        aerospike.unset_serializers()

    def setup_method(self, method):
        """
        Setup method
        """

        self.delete_keys = []

    def teardown_method(self, method):
        """
        Teardown method
        """
        for key in self.delete_keys:
            TestUserSerializer.client.remove(key)

    def test_put_with_float_data_user_serializer(self):

        #    Invoke put() for float data record with user serializer.

        if TestUserSerializer.skip_old_server is False:
            aerospike.set_serializer(serialize_function)
            aerospike.set_deserializer(deserialize_function)
        else:
            aerospike.set_serializer(serialize_function_old_server)
            aerospike.set_deserializer(deserialize_function_old_server)
        key = ("test", "demo", 1)

        rec = {"pi": 3.14}

        res = TestUserSerializer.client.put(key, rec, {}, {}, aerospike.SERIALIZER_USER)

        assert res == 0

        _, _, bins = TestUserSerializer.client.get(key)

        assert bins == {"pi": 3.14}

        self.delete_keys.append(key)

    def test_put_with_bool_data_user_serializer(self):
        """
        Invoke put() for bool data record with user serializer.
        """

        aerospike.set_serializer(serialize_function)
        aerospike.set_deserializer(deserialize_function)
        key = ("test", "demo", 1)

        rec = {"status": True}

        res = TestUserSerializer.client.put(key, rec, {}, {}, aerospike.SERIALIZER_USER)

        assert res == 0

        _, _, bins = TestUserSerializer.client.get(key)

        assert bins == {"status": True}

        self.delete_keys.append(key)

    """
    def test_put_with_object_data_user_serializer(self):

            #Invoke put() for object data record with user serializer.

        key = ( 'test', 'demo', 1 )

        obj1 = SomeClass()
        rec = {
            'object': obj1
        }

        res = TestUserSerializer.client.put( key, rec , {}, {},
            aerospike.SERIALIZER_USER)

        assert res == 0

        _, _, bins = TestUserSerializer.client.get( key )

        #assert bins == { 'object': True }
        assert bins == { 'object': obj1 }

        self.delete_keys.append( key )

    def test_put_with_object_data_python_serializer(self):

            Invoke put() for object data record with python serializer.

        key = ( 'test', 'demo', 1 )

        obj1 = SomeClass()
        rec = {
                "object" : obj1
        }

        res = TestUserSerializer.client.put( key, rec , {}, {},
            aerospike.SERIALIZER_PYTHON)

        assert res == 0
        _, _, bins = TestUserSerializer.client.get( key )

        assert bins == { 'object': True }

        self.delete_keys.append( key )
    """

    def test_put_with_float_data_user_serializer_none(self):
        """
        Invoke put() for float data record with user serializer.
        """

        try:
            aerospike.set_serializer(None)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Parameter must be a callable"

    def test_put_with_float_data_user_deserializer_none(self):
        """
        Invoke put() for float data record with user deserializer None.
        """

        if TestUserSerializer.skip_old_server is False:
            aerospike.set_serializer(serialize_function)
        else:
            aerospike.set_serializer(serialize_function_old_server)

        key = ("test", "demo", 1)

        rec = {"pi": 3.14}

        res = TestUserSerializer.client.put(key, rec, {}, {}, aerospike.SERIALIZER_USER)

        assert res == 0

        self.delete_keys.append(key)

        try:
            aerospike.set_deserializer(None)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Parameter must be a callable"

    def test_put_with_mixed_data_user_serializer(self):
        pytest.xfail(reason="Need Python 2/3 compatible bytearray for strings")
        """
            Invoke put() for mixed data record with user serializer.
        """

        key = ("test", "demo", 1)

        if TestUserSerializer.skip_old_server is False:
            aerospike.set_serializer(serialize_function)
            aerospike.set_deserializer(deserialize_function)
        else:
            aerospike.set_serializer(serialize_function_old_server)
            aerospike.set_deserializer(deserialize_function_old_server)
        rec = {
            "map": {"key": "asd';q;'1';", "pi": 3.14},
            "normal": 1234,
            "special": "!@#@#$QSDAsd;as",
            "list": ["nanslkdl", 1, bytes("asd;as[d'as;d", "utf-8")],
            "bytes": bytes("asd;as[d'as;d", "utf-8"),
            "nestedlist": ["nanslkdl", 1, bytes("asd;as[d'as;d", "utf-8"), [1, bytes("asd;as[d'as;d", "utf-8")]],
            "nestedmap": {"key": "asd';q;'1';", "pi": 3.14, "nest": {"pi1": 3.12, "t": 1}},
        }
        res = TestUserSerializer.client.put(key, rec, {}, {}, aerospike.SERIALIZER_USER)

        assert res == 0

        _, _, bins = TestUserSerializer.client.get(key)

        assert bins == {
            "map": {"key": "asd';q;'1';", "pi": 3.14},
            "normal": 1234,
            "special": "!@#@#$QSDAsd;as",
            "list": ["nanslkdl", 1, bytes("asd;as[d'as;d", "utf-8")],
            "bytes": bytes("asd;as[d'as;d", "utf-8"),
            "nestedlist": ["nanslkdl", 1, bytes("asd;as[d'as;d", "utf-8"), [1, bytes("asd;as[d'as;d", "utf-8")]],
            "nestedmap": {"key": "asd';q;'1';", "pi": 3.14, "nest": {"pi1": 3.12, "t": 1}},
        }

        self.delete_keys.append(key)

    def test_put_with_mixeddata_client_serializer_deserializer_with_spec_in_put(self):
        pytest.xfail(reason="Need Python 2/3 compatible bytearray for strings")

        #    Invoke put() for mixed data with class and instance serialziers
        #    with a specification in put. Client one is called

        method_config = TestBaseClass.get_connection_config()
        method_config["serialization"] = (client_serialize_function, client_deserialize_function)
        if method_config["user"] is None and method_config["password"] is None:
            client = aerospike.client(method_config).connect()
        else:
            client = aerospike.client(method_config).connect(method_config["user"], method_config["password"])
        aerospike.set_serializer(serialize_function)
        aerospike.set_deserializer(deserialize_function)
        key = ("test", "demo", 1)

        rec = {
            "map": {"key": "asd';q;'1';", "pi": 3},
            "normal": 1234,
            "special": "!@#@#$QSDAsd;as",
            "list": ["nanslkdl", 1, bytearray("asd;as[d'as;d")],
            "bytes": bytearray("asd;as[d'as;d"),
            "nestedlist": ["nanslkdl", 1, bytearray("asd;as[d'as;d"), [1, bytearray("asd;as[d'as;d")]],
            "nestedmap": {"key": "asd';q;'1';", "pi": 314, "nest": {"pi1": 312, "t": 1}},
        }

        res = client.put(key, rec, {}, {}, aerospike.SERIALIZER_USER)

        assert res == 0

        _, _, bins = client.get(key)

        assert bins == {
            "map": {"key": "asd';q;'1';", "pi": 3},
            "normal": 1234,
            "special": "!@#@#$QSDAsd;as",
            "list": ["nanslkdl", 1, bytearray("asd;as[d'as;d", "utf-8")],
            "bytes": bytearray("asd;as[d'as;d", "utf-8"),
            "nestedlist": [
                "nanslkdl",
                1,
                bytearray("asd;as[d'as;d", "utf-8"),
                [1, bytearray("asd;as[d'as;d", "utf-8")],
            ],
            "nestedmap": {"key": "asd';q;'1';", "pi": 314, "nest": {"pi1": 312, "t": 1}},
        }
        client.close()

        self.delete_keys.append(key)

    def test_put_with_mixeddata_client_serializer_deserializer_no_put_spec(self):
        pytest.xfail(reason="Need Python 2/3 compatible bytearray for strings")

        #    Invoke put() for mixed data with class and instance serialziers
        #    with no specification in put
        method_config = TestBaseClass.get_connection_config()
        method_config["serialization"] = (client_serialize_function, client_deserialize_function)
        if method_config["user"] is None and method_config["password"] is None:
            client = aerospike.client(method_config).connect()
        else:
            client = aerospike.client(method_config).connect(method_config["user"], method_config["password"])

        aerospike.set_serializer(serialize_function)
        aerospike.set_deserializer(deserialize_function)
        key = ("test", "demo", 1)

        rec = {
            "map": {"key": "asd';q;'1';", "pi": 3},
            "normal": 1234,
            "special": "!@#@#$QSDAsd;as",
            "list": ["nanslkdl", 1, bytes("asd;as[d'as;d", "utf-8")],
            "bytes": bytes("asd;as[d'as;d", "utf-8"),
            "nestedlist": ["nanslkdl", 1, bytes("asd;as[d'as;d", "utf-8"), [1, bytes("asd;as[d'as;d", "utf-8")]],
            "nestedmap": {"key": "asd';q;'1';", "pi": 3.14, "nest": {"pi1": 312, "t": 1}},
        }

        res = client.put(key, rec, {}, {})

        assert res == 0

        _, _, bins = client.get(key)

        assert bins == {
            "map": {"key": "asd';q;'1';", "pi": 3},
            "normal": 1234,
            "special": "!@#@#$QSDAsd;as",
            "list": ["nanslkdl", 1, "asd;as[d'as;d"],
            "bytes": "asd;as[d'as;d",
            "nestedlist": ["nanslkdl", 1, "asd;as[d'as;d", [1, "asd;as[d'as;d"]],
            "nestedmap": {"key": "asd';q;'1';", "pi": 3.14, "nest": {"pi1": 312, "t": 1}},
        }
        client.close()
        self.delete_keys.append(key)



# Insert 3 records
keys = [("test", "demo", f"employee{i}") for i in range(1, 4)]
bins = [
    {"id": 100, "balance": 200},
    {"id": 101, "balance": 400},
    {"id": 102, "balance": 300}
]
for key, bin in zip(keys, bins):
    client.put(key, bin)

batchRecords = client.batch_remove(keys)

# A result of 0 means success
print(batchRecords.result)
# 0
for batchRecord in batchRecords.batch_records:
    print(batchRecord.result)
    print(batchRecord.record)
# 0: (('test', 'demo', 'employee1', bytearray(b'...')), {'ttl': 4294967295, 'gen': 0}, {})
# 0: (('test', 'demo', 'employee2', bytearray(b'...')), {'ttl': 4294967295, 'gen': 0}, {})
# 0: (('test', 'demo', 'employee3', bytearray(b'...')), {'ttl': 4294967295, 'gen': 0}, {})



# -*- coding: utf-8 -*-
import pytest
from aerospike import exception as e
from .test_base_class import TestBaseClass

import aerospike


class TestUdfList(object):
    def setup_class(cls):
        """
        Setup class
        """
        cls.client = TestBaseClass.get_new_connection()
        cls.client.udf_put("example.lua", 0, {})

    def teardown_class(cls):
        """
        Remove the UDF we added, and close the connection
        """
        cls.client.udf_remove("example.lua")
        cls.client.close()

    def test_udf_list_without_any_parameters(self):
        """
        Test successful call to udf_list with no parameters
        """
        udf_list = self.client.udf_list()
        assert type(udf_list) is list
        present = False
        for udf in udf_list:
            if "example.lua" == udf["name"]:
                present = True

        assert present

    def test_udf_list_with_proper_parameters(self):
        """
        Test successful call to udf_list with a policy supplied
        """
        policy = {"timeout": 180000}
        udf_list = self.client.udf_list(policy)

        present = False
        for udf in udf_list:
            if "example.lua" == udf["name"]:
                present = True

        assert present

    def test_udf_list_with_none_type_policy(self):
        """
        Test calling udf_list with a None type policy
        """
        policy = None
        udf_list = self.client.udf_list(policy)

        present = False
        for udf in udf_list:
            if "example.lua" == udf["name"]:
                present = True

        assert present

    def test_udf_list_with_timeout_policy(self):
        """
        Test successful call to udf_list with a timeout of 1s
        """
        policy = {"timeout": 180000}

        udf_list = self.client.udf_list(policy)

        present = False
        for udf in udf_list:
            if "example.lua" == udf["name"]:
                present = True

        assert present

    def test_udf_list_with_invalid_timeout_policy_value(self):
        """
        Test to verify error raised by passing an invalid timeout value
        to udf_list
        """
        policy = {"timeout": 0.1}

        with pytest.raises(e.ParamError) as param_error:
            self.client.udf_list(policy)

        assert param_error.value.code == -2
        assert param_error.value.msg == "timeout is invalid"

    def test_udf_list_with_extra_arg(self):
        policy = {}
        with pytest.raises(TypeError):
            self.client.udf_list(policy, "extra_parameter")

    @pytest.mark.parametrize("policy", [False, 1000, [], (), "timeout:5000"])
    def test_udf_list_with_wrong_policy_type(self, policy):
        """
        Tests for behavior caused by calling udf_list with
        the wrong type of policy
        """
        with pytest.raises(e.ParamError):
            self.client.udf_list(policy)

    def test_udf_list_with_proper_parameters_without_connection(self):
        """
        Test to verify error raised by trying to call udf_list without
        first calling connect
        """
        config = TestBaseClass.get_connection_config()

        client1 = aerospike.client(config)
        client1.close()

        policy = {"timeout": 180000}

        with pytest.raises(e.ClusterError) as cluster_error:
            client1.udf_list(policy)

        assert cluster_error.value.code == 11
        assert cluster_error.value.msg == "No connection to aerospike cluster"



import aerospike
from aerospike_helpers import cdt_ctx

config = {'hosts': [('127.0.0.1', 3000)]}
client = aerospike.client(config)

ctxs = [cdt_ctx.cdt_ctx_list_index(0)]
ctxs_base64 = client.get_cdtctx_base64(ctxs)
print("Base64 encoding of ctxs:", ctxs_base64)

client.close()



# -*- coding: utf-8 -*-

import pytest
from .as_status_codes import AerospikeStatus
from .udf_helpers import wait_for_udf_removal, wait_for_udf_to_exist
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


@pytest.mark.usefixtures("as_connection")
class TestUdfPut(TestBaseClass):
    def setup_class(cls):
        cls.udf_name = "example.lua"

    def teardown_method(self, method):
        """
        Teardown method
        """
        udf_list = self.as_connection.udf_list()
        for udf in udf_list:
            if udf["name"] == self.udf_name:
                self.as_connection.udf_remove(self.udf_name)
                wait_for_udf_removal(self.as_connection, self.udf_name)

    def test_udf_put_with_proper_parameters_no_policy(self):
        """
        Test to verify execution of udf_put with proper parameters,
        and an empty policy dict
        """

        filename = self.udf_name
        udf_type = 0

        status = self.as_connection.udf_put(filename, udf_type)

        assert status == 0
        udf_list = self.as_connection.udf_list({})

        present = False
        for udf in udf_list:
            if self.udf_name == udf["name"]:
                present = True
                break

        assert present

    def test_udf_put_with_proper_timeout_policy_value(self):
        """
        Test that calling udf_put with the proper arguments will add the udf
        to the server
        """

        policy = {"timeout": 180000}
        filename = self.udf_name
        udf_type = 0

        status = self.as_connection.udf_put(filename, udf_type, policy)

        assert status == 0

        udf_list = self.as_connection.udf_list({})

        present = False
        for udf in udf_list:
            if udf["name"] == filename:
                present = True
                break

        assert present

    def test_udf_put_with_filename_unicode(self):

        policy = {}
        filename = "example.lua"
        udf_type = 0

        status = self.as_connection.udf_put(filename, udf_type, policy)

        assert status == 0

        # wait for the udf to propagate to the server
        wait_for_udf_to_exist(self.as_connection, filename)
        udf_list = self.as_connection.udf_list({})

        present = False
        for udf in udf_list:
            if "example.lua" == udf["name"]:
                present = True

        assert present

    def test_udf_put_empty_script_file(self):

        policy = {}
        filename = "empty.lua"
        udf_type = 0
        with pytest.raises(e.LuaFileNotFound):
            self.as_connection.udf_put(filename, udf_type, policy)

    def test_udf_put_with_filename_too_long(self):

        policy = {}
        filename = "a" * 510 + ".lua"
        udf_type = 0
        with pytest.raises(e.ParamError):
            self.as_connection.udf_put(filename, udf_type, policy)

    def test_udf_put_with_empty_filename(self):

        policy = {}
        filename = ""
        udf_type = 0
        with pytest.raises(e.ParamError):
            self.as_connection.udf_put(filename, udf_type, policy)

    def test_udf_put_with_empty_filename_beginning_with_slash(self):

        policy = {}
        filename = "/"
        udf_type = 0
        with pytest.raises(e.ParamError):
            self.as_connection.udf_put(filename, udf_type, policy)

    def test_udf_put_with_proper_parameters_without_connection(self):

        policy = {}
        filename = self.udf_name
        udf_type = 0

        config = TestBaseClass.get_connection_config()

        client1 = aerospike.client(config)
        client1.close()

        with pytest.raises(e.ClusterError) as err_info:
            client1.udf_put(filename, udf_type, policy)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_CLUSTER_ERROR

    def test_udf_put_with_invalid_timeout_policy_value(self):
        """
        Test that invalid timeout policy causes an error on udf put
        """
        policy = {"timeout": 0.1}
        filename = self.udf_name
        udf_type = 0

        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.udf_put(filename, udf_type, policy)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_udf_put_without_parameters(self):
        """
        Test that calling udf_put without parameters raises
        an error
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.udf_put()

        assert "argument 'filename' (pos 1)" in str(typeError.value)

    def test_udf_put_with_non_existent_filename(self):
        """
        Test that an error is raised when an invalid filename
        is given to udf_put
        """
        policy = {}
        filename = "somefile_that_does_not_exist"
        udf_type = 0

        with pytest.raises(e.LuaFileNotFound) as err_info:
            self.as_connection.udf_put(filename, udf_type, policy)

        assert err_info.value.code == AerospikeStatus.LUA_FILE_NOT_FOUND

    def test_udf_put_with_non_lua_udf_type_and_lua_script_file(self):
        """
        Test to verify that an invalid udf_type causes an error
        """
        policy = {"timeout": 180000}
        filename = self.udf_name
        udf_type = 1

        with pytest.raises(e.ClientError) as err_info:
            self.as_connection.udf_put(filename, udf_type, policy)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_udf_put_with_all_none_parameters(self):
        """
        Test to verify that calling udf_put with all 3 parameters as None
        will raise an error
        """

        with pytest.raises(TypeError) as exception:
            self.as_connection.udf_put(None, None, None)

        assert "an integer is required" or "cannot be interpreted as an integer" in str(exception.value)

    @pytest.mark.parametrize(
        "filename, ftype, policy",
        (
            (1, 0, {}),
            (None, 0, {}),
            ((), 0, {}),
            ("example.lua", "0", {}),
            ("example.lua", (), {}),
            ("example.lua", None, {}),
            ("example.lua", 0, []),
            ("example.lua", 0, "policy"),
            ("example.lua", 0, 5),
        ),
    )
    def test_udf_put_invalid_arg_types(self, filename, ftype, policy):
        """
        Incorrect type for second argument raise a type error,
        the others cause a param error
        """
        with pytest.raises((e.ParamError, TypeError)):
            self.as_connection.udf_put(filename, ftype, policy)

    def test_udf_put_with_extra_arg(self):
        policy = {}
        with pytest.raises(TypeError):
            self.as_connection.udf_put(self.udf_name, 1, policy, "extra_arg")



from aerospike import predicates

query.select('score')
query.where(predicates.equals('score', 100))

records = query.results()
# Matches one record
print(records)
# [(('test', 'demo', None, bytearray(b'...')), {'ttl': 2592000, 'gen': 1}, {'score': 100})]



# -*- coding: utf-8 -*-

import pytest
import time
import os

from contextlib import contextmanager
from .test_base_class import TestBaseClass

test_memleak = int(os.environ.get("TEST_MEMLEAK", 0))
if test_memleak != 1:
    if pytest.__version__ < "3.0.0":
        pytest.skip("Skip memleak tests")
    else:
        pytestmark = pytest.mark.skip

test_memleak_loop = int(os.environ.get("TEST_MEMLEAK_LOOP", 1))


@contextmanager
def open_as_connection(config):
    """
    Context manager to let us open aerospike connections with
    specified config
    """
    as_connection = TestBaseClass.get_new_connection(config)

    # Connection is setup, so yield it
    yield as_connection

    # close the connection
    as_connection.close()


# adds cls.connection_config to this class
@pytest.mark.usefixtures("connection_config")
class TestConnectLeak(object):
    def test_connect_leak(self):
        """
        Invoke connect() with positive parameters.
        """
        # first_ref_count = sys.gettotalrefcount()
        # last_ref_count = first_ref_count
        # print("-----start gettotalrefcount: " + str(first_ref_count))

        config = self.connection_config.copy()
        i = 0
        while i < test_memleak_loop:
            with open_as_connection(config) as client:
                assert client is not None
                assert client.is_connected()
                time.sleep(0.1)
            i = i + 1
        # last_ref_count = sys.gettotalrefcount()
        # print("-----outstanding gettotalrefcount: " + str(last_ref_count-first_ref_count))



# data used in fixtures for invalid arguments

INVALID_KEYS = [
    None,
    (None, "demo", 1),
    (1, "demo", 1),
    (False, "demo", 1),
    ([], "demo", 1),
    ("test", 1, 1),
    ("test", False, 1),
    ("test", [], 1),
    ("test", "demo", None),
    ("test", "demo", []),
    ("test", "demo", {}),
]

INVALID_BINS = [
    None,
    False,
    [],
]



# -*- coding: utf-8 -*-

import pytest
import time
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


class TestSetWhitelist(TestBaseClass):

    config = TestBaseClass.get_connection_config()

    pytestmark = pytest.mark.skipif(
        not TestBaseClass.auth_in_use(), reason="No user specified, may be not secured cluster."
    )

    def setup_method(self, method):
        """
        Setup method
        """
        config = self.config
        self.client = aerospike.client(config).connect(config["user"], config["password"])
        usr_sys_admin_privs = [{"code": aerospike.PRIV_USER_ADMIN}, {"code": aerospike.PRIV_SYS_ADMIN}]
        whitelist = ["127.0.0.1"]
        try:
            self.client.admin_drop_role("usr-sys-admin-test")
            time.sleep(2)
        except Exception:
            pass
        self.client.admin_create_role("usr-sys-admin-test", usr_sys_admin_privs, whitelist=whitelist)

        self.delete_users = []
        time.sleep(1)

    def teardown_method(self, method):
        """
        Teardown method
        """
        try:
            self.client.admin_drop_role("usr-sys-admin-test")
        except Exception:
            pass
        self.client.close()

    def test_admin_set_whitelist_no_parameters(self):
        """
        Set whitelist with no parameters.
        """
        with pytest.raises(TypeError):
            self.client.admin_set_whitelist()

    def test_admin_set_whitelist_empty_whitelist_positive(self):
        """
        Set whitelist with no whitelist. (will reset whitelist on a role with existing whitelist)
        """
        self.client.admin_set_whitelist(role="usr-sys-admin-test", whitelist=[])
        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")
        assert roles == {
            "privileges": [{"ns": "", "set": "", "code": 0}, {"ns": "", "set": "", "code": 1}],
            "whitelist": [],
            "read_quota": 0,
            "write_quota": 0,
        }

    def test_admin_set_whitelist_none_whitelist_positive(self):
        """
        Set whitelist with no whitelist. (will reset whitelist on a role with existing whitelist)
        """
        self.client.admin_set_whitelist(role="usr-sys-admin-test", whitelist=None)
        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")
        assert roles == {
            "privileges": [{"ns": "", "set": "", "code": 0}, {"ns": "", "set": "", "code": 1}],
            "whitelist": [],
            "read_quota": 0,
            "write_quota": 0,
        }

    def test_admin_set_whitelist_one_whitelist_positive(self):
        """
        Set whitelist with whitelist.
        """
        self.client.admin_set_whitelist(role="usr-sys-admin-test", whitelist=["10.0.2.0/24", "127.0.0.1", "127.0.0.2"])
        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")
        assert roles == {
            "privileges": [{"ns": "", "set": "", "code": 0}, {"ns": "", "set": "", "code": 1}],
            "whitelist": ["10.0.2.0/24", "127.0.0.1", "127.0.0.2"],
            "read_quota": 0,
            "write_quota": 0,
        }

    def test_admin_set_quota_empty_positive(self):
        """
        Set whitelist positive
        """
        self.client.admin_set_whitelist(role="usr-sys-admin-test", whitelist=[])
        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")
        assert roles == {
            "privileges": [{"ns": "", "set": "", "code": 0}, {"ns": "", "set": "", "code": 1}],
            "whitelist": [],
            "read_quota": 0,
            "write_quota": 0,
        }

    def test_admin_set_whitelist_positive_with_policy(self):
        """
        Set whitelist positive policy
        """
        self.client.admin_set_whitelist(
            role="usr-sys-admin-test", whitelist=["10.0.2.0/24", "127.0.0.1"], policy={"timeout": 1000}
        )
        time.sleep(1)
        roles = self.client.admin_get_role("usr-sys-admin-test")
        assert roles == {
            "privileges": [{"ns": "", "set": "", "code": 0}, {"ns": "", "set": "", "code": 1}],
            "whitelist": ["10.0.2.0/24", "127.0.0.1"],
            "read_quota": 0,
            "write_quota": 0,
        }

    def test_admin_set_whitelist_incorrect_role_name(self):
        """
        Incorrect role name
        """
        try:
            self.client.admin_set_whitelist(role="bad-role-name", whitelist=["10.0.2.0/24"], policy={"timeout": 1000})

        except e.InvalidRole as exception:
            assert exception.code == 70
            assert exception.msg == "AEROSPIKE_INVALID_ROLE"

    def test_admin_set_whitelist_incorrect_role_type(self):
        """
        Incorrect role type
        """
        try:
            self.client.admin_set_whitelist(role=None, whitelist=["10.0.2.0/24"], policy={"timeout": 1000})

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Role name should be a string."

    def test_admin_set_whitelist_incorrect_whitelist(self):
        """
        Incorrect role name
        """
        try:
            self.client.admin_set_whitelist(role="usr-sys-admin-test", whitelist=["bad_IP"], policy={"timeout": 1000})
        except e.InvalidWhitelist as exception:
            assert exception.code == 73
            assert exception.msg == "AEROSPIKE_INVALID_WHITELIST"

    def test_admin_set_whitelist_incorrect_whitelist_type(self):
        """
        Incorrect role type
        """
        try:
            self.client.admin_set_whitelist(role="usr-sys-admin-test", whitelist=None, policy={"timeout": 1000})

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Whitelist must be a list of IP strings."

    def test_admin_set_whitelist_forbiden_host(self):
        """
        Forbiden host
        """
        self.client.admin_set_whitelist(role="usr-sys-admin-test", whitelist=["123.4.5.6"], policy={"timeout": 1000})

        self.client.admin_create_user("test_whitelist_user", "123", ["usr-sys-admin-test"])

        config = TestBaseClass.get_connection_config()
        new_client = aerospike.client(config).connect(config["user"], config["password"])
        try:
            new_client.connect("test_whitelist_user", "123")
        except e.NotWhitelisted as exception:
            assert exception.code == 82
            assert exception.msg == "Failed to connect"
        finally:
            self.client.admin_drop_user("test_whitelist_user")



# -*- coding: utf-8 -*-
##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################


from __future__ import print_function
import aerospike
import json
import sys

from optparse import OptionParser

##########################################################################
# Option Parsing
##########################################################################

usage = "usage: %prog [options] key module function [args...]"

optparser = OptionParser(usage=usage, add_help_option=False)

optparser.add_option(
    "--help", dest="help", action="store_true",
    help="Displays this message.")

optparser.add_option(
    "-U", "--username", dest="username", type="string", metavar="<USERNAME>",
    help="Username to connect to database.")

optparser.add_option(
    "-P", "--password", dest="password", type="string", metavar="<PASSWORD>",
    help="Password to connect to database.")

optparser.add_option(
    "-h", "--host", dest="host", type="string", default="127.0.0.1", metavar="<ADDRESS>",
    help="Address of Aerospike server.")

optparser.add_option(
    "-p", "--port", dest="port", type="int", default=3000, metavar="<PORT>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-n", "--namespace", dest="namespace", type="string", default="test", metavar="<NS>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-s", "--set", dest="set", type="string", default="demo", metavar="<SET>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "--gen", dest="gen", type="int", default=None, metavar="<GEN>",
    help="Generation of the record being written.")

optparser.add_option(
    "--ttl", dest="ttl", type="int", default=None, metavar="<TTL>",
    help="TTL of the record being written.")


(options, args) = optparser.parse_args()

if options.help:
    optparser.print_help()
    print()
    sys.exit(1)

if len(args) < 3:
    optparser.print_help()
    print()
    sys.exit(1)

##########################################################################
# Client Configuration
##########################################################################

config = {
    'hosts': [(options.host, options.port)]
}

##########################################################################
# Application
##########################################################################

exitCode = 0


def parse_arg(s):
    try:
        return json.loads(s)
    except ValueError:
        return s

try:

    # ----------------------------------------------------------------------------
    # Connect to Cluster
    # ----------------------------------------------------------------------------

    client = aerospike.client(config).connect(
        options.username, options.password)

    # ----------------------------------------------------------------------------
    # Perform Operation
    # ----------------------------------------------------------------------------

    try:

        namespace = options.namespace if options.namespace and options.namespace != 'None' else None
        set = options.set if options.set and options.set != 'None' else None

        args.reverse()
        key = args.pop()
        module = args.pop()
        function = args.pop()

        # invoke operation
        args.reverse()
        argl = list(map(parse_arg, args))
        res = client.apply((namespace, set, key), module, function, argl)

        print(res)
        print("---")
        print("OK, 1 UDF applied.")

    except Exception as e:
        print("error: {0}".format(e), file=sys.stderr)
        exitCode = 2

    # ----------------------------------------------------------------------------
    # Close Connection to Cluster
    # ----------------------------------------------------------------------------

    client.close()

except Exception as eargs:
    print("error: {0}".format(eargs), file=sys.stderr)
    exitCode = 3

##########################################################################
# Exit
##########################################################################

sys.exit(exitCode)
